#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import numpy as np
import torch
from ax.core import Metric, Objective, OptimizationConfig
from ax.core.observation import ObservationData, ObservationFeatures
from ax.modelbridge.pairwise import (
    _binary_pref_to_comp_pair,
    _consolidate_comparisons,
    PairwiseModelBridge,
)
from ax.models.torch.botorch_modular.model import BoTorchModel
from ax.models.torch.botorch_modular.surrogate import Surrogate
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import checked_cast
from ax.utils.testing.preference_stubs import get_pbo_experiment
from botorch.acquisition.monte_carlo import qNoisyExpectedImprovement
from botorch.acquisition.preference import (
    AnalyticExpectedUtilityOfBestOption,
    qExpectedUtilityOfBestOption,
)
from botorch.models.pairwise_gp import PairwiseGP, PairwiseLaplaceMarginalLogLikelihood
from botorch.models.transforms.input import Normalize
from botorch.utils.datasets import RankingDataset


class PairwiseModelBridgeTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        experiment = get_pbo_experiment()
        self.experiment = experiment
        self.data = experiment.lookup_data()

    def test_PairwiseModelBridge(self) -> None:
        surrogate = Surrogate(
            botorch_model_class=PairwiseGP,
            mll_class=PairwiseLaplaceMarginalLogLikelihood,
            input_transform_classes=[Normalize],
            input_transform_options={
                "Normalize": {"d": len(self.experiment.parameters)}
            },
        )

        cases = [
            (qNoisyExpectedImprovement, None, 3),
            (qExpectedUtilityOfBestOption, None, 3),
            (
                AnalyticExpectedUtilityOfBestOption,
                # Analytic Acqfs do not support pending points and sequential opt
                {"optimizer_kwargs": {"sequential": False}},
                2,  # analytic EUBO only supports n=2
            ),
        ]
        for botorch_acqf_class, model_gen_options, n in cases:
            pmb = PairwiseModelBridge(
                experiment=self.experiment,
                search_space=self.experiment.search_space,
                data=self.data,
                model=BoTorchModel(
                    botorch_acqf_class=botorch_acqf_class,
                    surrogate=surrogate,
                ),
                transforms=[],
                optimization_config=OptimizationConfig(
                    Objective(
                        Metric(Keys.PAIRWISE_PREFERENCE_QUERY.value), minimize=False
                    )
                ),
                fit_tracking_metrics=False,
            )
            # Can generate candidates correctly
            # pyre-ignore: Incompatible parameter type [6]
            generator_run = pmb.gen(n=n, model_gen_options=model_gen_options)
            self.assertEqual(len(generator_run.arms), n)

        observation_data = [
            ObservationData(
                metric_names=[Keys.PAIRWISE_PREFERENCE_QUERY.value],
                means=np.array([0]),
                covariance=np.array([[np.nan]]),
            ),
            ObservationData(
                metric_names=[Keys.PAIRWISE_PREFERENCE_QUERY.value],
                means=np.array([1]),
                covariance=np.array([[np.nan]]),
            ),
        ]
        observation_features = [
            ObservationFeatures(parameters={"x1": 0.1, "x2": 0.2}, trial_index=0),
            ObservationFeatures(parameters={"x1": 0.3, "x2": 0.4}, trial_index=0),
        ]
        observation_features_with_metadata = [
            ObservationFeatures(parameters={"x1": 0.1, "x2": 0.2}, trial_index=0),
            ObservationFeatures(
                parameters={"x1": 0.3, "x2": 0.4},
                trial_index=0,
                metadata={"metadata_key": "metadata_val"},
            ),
        ]
        parameter_names = list(self.experiment.parameters.keys())
        outcomes = [checked_cast(str, Keys.PAIRWISE_PREFERENCE_QUERY.value)]

        datasets, _, candidate_metadata = pmb._convert_observations(
            observation_data=observation_data,
            observation_features=observation_features,
            outcomes=outcomes,
            parameters=parameter_names,
            search_space_digest=None,
        )
        self.assertTrue(len(datasets) == 1)
        self.assertTrue(isinstance(datasets[0], RankingDataset))
        self.assertTrue(candidate_metadata is None)

        datasets, _, candidate_metadata = pmb._convert_observations(
            observation_data=observation_data,
            observation_features=observation_features_with_metadata,
            outcomes=outcomes,
            parameters=parameter_names,
            search_space_digest=None,
        )
        self.assertTrue(len(datasets) == 1)
        self.assertTrue(isinstance(datasets[0], RankingDataset))
        self.assertTrue(candidate_metadata is not None)

        # Test individual helper methods
        X = torch.tensor(
            [[1.0, 2.0, 3.0], [2.0, 3.0, 4.0], [1.0, 2.0, 3.0], [2.1, 3.1, 4.1]]
        )
        Y = torch.tensor([[1, 0, 0, 1]])
        expected_X = torch.tensor([[1.0, 2.0, 3.0], [2.0, 3.0, 4.0], [2.1, 3.1, 4.1]])
        ordered_Y = torch.tensor([[0, 1], [3, 2]])
        expected_Y = torch.tensor([[0, 1], [2, 0]])

        # `_binary_pref_to_comp_pair`.
        comp_pair_Y = _binary_pref_to_comp_pair(Y=Y)
        self.assertTrue(torch.equal(comp_pair_Y, ordered_Y))

        # test `_binary_pref_to_comp_pair` with invalid data
        bad_Y = torch.tensor([[1, 1, 0, 0]])
        with self.assertRaises(ValueError):
            _binary_pref_to_comp_pair(Y=bad_Y)

        # `_consolidate_comparisons`.
        consolidated_X, consolidated_Y = _consolidate_comparisons(X=X, Y=comp_pair_Y)
        self.assertTrue(torch.equal(consolidated_X, expected_X))
        self.assertTrue(torch.equal(consolidated_Y, expected_Y))

        with self.assertRaises(ValueError):
            _consolidate_comparisons(
                X=X.expand(2, *X.shape), Y=comp_pair_Y.expand(2, *comp_pair_Y.shape)
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.utils.common.result import Err, Ok, Result, UnwrapError
from ax.utils.common.testutils import TestCase


class ResultTest(TestCase):
    def setUp(self) -> None:
        super().setUp()

        def safeDivide(a: float, b: float) -> Result[float, str]:
            if b == 0:
                return Err("yikes")

            return Ok(a / b)

        self.ok = safeDivide(0, 2)
        self.err = safeDivide(0, 0)

    def test_eq(self) -> None:
        self.assertEqual(self.ok, Ok(0))
        self.assertEqual(self.err, Err("yikes"))

    def test_repr(self) -> None:
        self.assertEqual(str(self.ok), "Ok(0.0)")
        self.assertEqual(str(self.err), "Err(yikes)")

    def test_is(self) -> None:
        self.assertTrue(self.ok.is_ok())
        self.assertFalse(self.ok.is_err())

        self.assertFalse(self.err.is_ok())
        self.assertTrue(self.err.is_err())

    def test_map(self) -> None:
        def f(n: int) -> int:
            return n + 1

        def g(val: str) -> int:
            return len(val)

        def h() -> int:
            return -1

        self.assertEqual(self.ok.map(op=f), Ok(1))
        self.assertEqual(self.ok.map_err(op=g), Ok(0))
        self.assertEqual(self.ok.map_or(default="foo", op=f), 1)
        self.assertEqual(self.ok.map_or_else(default_op=h, op=f), 1)

        self.assertEqual(self.err.map(op=f), Err("yikes"))
        self.assertEqual(self.err.map_err(op=g), Err(5))
        self.assertEqual(self.err.map_or(default="foo", op=f), "foo")
        self.assertEqual(self.err.map_or_else(default_op=h, op=f), -1)

    def test_unwrap(self) -> None:
        self.assertEqual(self.ok.unwrap(), 0)
        with self.assertRaises(UnwrapError):
            self.ok.unwrap_err()
        self.assertEqual(self.ok.unwrap_or(1), 0)
        self.assertEqual(self.ok.unwrap_or_else(1), 0)

        with self.assertRaises(UnwrapError):
            self.err.unwrap()
        self.assertEqual(self.err.unwrap_err(), "yikes")
        self.assertEqual(self.err.unwrap_or(1), 1)
        self.assertEqual(self.err.unwrap_or_else(lambda s: 2), 2)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from ax.core.base_trial import BaseTrial
from ax.core.data import Data
from ax.core.metric import Metric, MetricFetchE, MetricFetchResult
from ax.utils.common.result import Err, Ok
from ax.utils.common.typeutils import not_none


class DictLookupMetric(Metric):
    """A metric defined by a dictionary mapping parameter values to the
    corresponding metric values.

    This provides an option to add normal noise with mean 0 and mean_sd scale
    to the given metric values.
    """

    def __init__(
        self,
        name: str,
        param_names: List[str],
        lookup_dict: Dict[Tuple[Union[str, float, int, bool], ...], float],
        noise_sd: Optional[float] = 0.0,
        lower_is_better: Optional[bool] = None,
    ) -> None:
        """Metric is computed via a dictionary look up using a tuple of
        parameter values, constructed based on the ordering of parameter
        names given in `param_names`.

        Args:
            name: Name of the metric.
            param_names: An ordered list of names of parameters to be used
                to construct the dictionary key.
            lookup_dict: A dictionary mapping a tuple of parameter values to
                the metric values.
            noise_sd: Scale of normal noise added to the function result. If
                None, interpret the function as noisy with unknown noise level.
            lower_is_better: Flag for metrics which should be minimized.
        """
        self.param_names = param_names
        self.lookup_dict = lookup_dict
        self.noise_sd = noise_sd
        super().__init__(name=name, lower_is_better=lower_is_better)

    @classmethod
    def is_available_while_running(cls) -> bool:
        return True

    def clone(self) -> DictLookupMetric:
        return self.__class__(
            name=self._name,
            param_names=self.param_names,
            lookup_dict=self.lookup_dict,
            noise_sd=self.noise_sd,
            lower_is_better=self.lower_is_better,
        )

    def fetch_trial_data(self, trial: BaseTrial, **kwargs: Any) -> MetricFetchResult:
        try:
            noise_sd = self.noise_sd
            arm_names = []
            mean = []
            for name, arm in trial.arms_by_name.items():
                arm_names.append(name)
                lookup_key = tuple(
                    not_none(arm.parameters[p]) for p in self.param_names
                )
                try:
                    val = self.lookup_dict[lookup_key]
                except KeyError:
                    raise KeyError(
                        "Got a KeyError while attempting to retrieve the "
                        f"parameterization {arm.parameters} from the lookup dict. "
                        f"This parameterization corresponds to {lookup_key=}."
                    )
                if noise_sd:
                    val = val + noise_sd * np.random.randn()
                mean.append(val)
            # Indicate unknown noise level in data.
            if noise_sd is None:
                noise_sd = float("nan")
            df = pd.DataFrame(
                {
                    "arm_name": arm_names,
                    "metric_name": self.name,
                    "mean": mean,
                    "sem": noise_sd,
                    "trial_index": trial.index,
                }
            )

            return Ok(value=Data(df=df))

        except Exception as e:
            return Err(
                MetricFetchE(message=f"Failed to fetch {self.name}", exception=e)
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import argparse
import logging
import os
import time
import warnings

import torch
from pytorch_lightning import LightningModule, loggers as pl_loggers, Trainer
from torch import nn
from torch.nn import functional as F
from torch.utils.data import DataLoader
from torchmetrics.functional.classification.accuracy import multiclass_accuracy
from torchvision import transforms
from torchvision.datasets import MNIST

warnings.filterwarnings("ignore")  # Disable data logger warnings
logging.getLogger("pytorch_lightning").setLevel(logging.ERROR)  # Disable GPU/TPU prints


def parse_args():
    parser = argparse.ArgumentParser(description="train mnist")
    parser.add_argument(
        "--log_path",
        type=str,
        required=True,
        help="dir to place tensorboard logs from all trials",
    )
    parser.add_argument(
        "--hidden_size_1", type=int, required=True, help="hidden size layer 1"
    )
    parser.add_argument(
        "--hidden_size_2", type=int, required=True, help="hidden size layer 2"
    )
    parser.add_argument(
        "--learning_rate", type=float, required=True, help="learning rate"
    )
    parser.add_argument("--epochs", type=int, required=True, help="number of epochs")
    parser.add_argument(
        "--dropout", type=float, required=True, help="dropout probability"
    )
    parser.add_argument("--batch_size", type=int, required=True, help="batch size")
    return parser.parse_args()


args = parse_args()

PATH_DATASETS = os.environ.get("PATH_DATASETS", ".")


class MnistModel(LightningModule):
    def __init__(self):
        super().__init__()

        # Tunable parameters
        self.hidden_size_1 = args.hidden_size_1
        self.hidden_size_2 = args.hidden_size_2
        self.learning_rate = args.learning_rate
        self.dropout = args.dropout
        self.batch_size = args.batch_size

        # Set class attributes
        self.data_dir = PATH_DATASETS

        # Hardcode some dataset specific attributes
        self.num_classes = 10
        self.dims = (1, 28, 28)
        channels, width, height = self.dims
        self.transform = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize((0.1307,), (0.3081,)),
            ]
        )

        # Create a PyTorch model
        layers = [nn.Flatten()]
        width = channels * width * height
        hidden_layers = [self.hidden_size_1, self.hidden_size_2]
        num_params = 0
        for hidden_size in hidden_layers:
            if hidden_size > 0:
                layers.append(nn.Linear(width, hidden_size))
                layers.append(nn.ReLU())
                layers.append(nn.Dropout(self.dropout))
                num_params += width * hidden_size
                width = hidden_size
        layers.append(nn.Linear(width, self.num_classes))
        num_params += width * self.num_classes

        # Save the model and parameter counts
        self.num_params = num_params
        self.model = nn.Sequential(*layers)  # No need to use Relu for the last layer

    def forward(self, x):
        x = self.model(x)
        return F.log_softmax(x, dim=1)

    def training_step(self, batch, batch_idx):
        x, y = batch
        logits = self(x)
        loss = F.nll_loss(logits, y)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        logits = self(x)
        loss = F.nll_loss(logits, y)
        preds = torch.argmax(logits, dim=1)
        acc = multiclass_accuracy(preds, y, num_classes=self.num_classes)
        self.log("val_acc", acc, prog_bar=False)
        return loss

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer

    def prepare_data(self):
        MNIST(self.data_dir, train=True, download=True)
        MNIST(self.data_dir, train=False, download=True)

    def setup(self, stage=None):
        self.mnist_train = MNIST(self.data_dir, train=True, transform=self.transform)
        self.mnist_val = MNIST(self.data_dir, train=False, transform=self.transform)

    def train_dataloader(self):
        return DataLoader(self.mnist_train, batch_size=self.batch_size)

    def val_dataloader(self):
        return DataLoader(self.mnist_val, batch_size=self.batch_size)


def run_training_job():

    mnist_model = MnistModel()

    # Initialize a trainer
    logger = pl_loggers.TensorBoardLogger(args.log_path)
    trainer = Trainer(
        logger=logger,
        log_every_n_steps=1,
        max_epochs=args.epochs,
        enable_progress_bar=False,
        deterministic=True,  # Do we want a bit of noise?
        default_root_dir=args.log_path,
    )
    logger.save()

    print(f"Logging to path: {args.log_path}.")

    # Train the model and log time
    start = time.time()
    trainer.fit(model=mnist_model)
    end = time.time()
    train_time = end - start

    # Compute the validation accuracy
    val_accuracy = trainer.validate()[0]["val_acc"]

    # Log the number of model parameters
    num_params = trainer.model.num_params

    # Print outputs
    print(
        f"train time: {train_time}, val acc: {val_accuracy}, num_params: {num_params}"
    )


if __name__ == "__main__":
    run_training_job()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy
from math import sqrt

import numpy as np
from ax.core.metric import Metric
from ax.core.objective import Objective
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.optimization_config import OptimizationConfig
from ax.core.outcome_constraint import OutcomeConstraint
from ax.core.parameter import ChoiceParameter, ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.core.types import ComparisonOp
from ax.modelbridge.transforms.stratified_standardize_y import StratifiedStandardizeY
from ax.modelbridge.transforms.tests.test_standardize_y_transform import osd_allclose
from ax.utils.common.testutils import TestCase


class StratifiedStandardizeYTransformTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.obsd1 = ObservationData(
            metric_names=["m1", "m2", "m2"],
            means=np.array([1.0, 2.0, 8.0]),
            covariance=np.array([[1.0, 0.2, 0.4], [0.2, 2.0, 0.8], [0.4, 0.8, 3.0]]),
        )
        self.obsd2 = ObservationData(
            metric_names=["m1", "m1", "m2", "m2"],
            means=np.array([1.0, 5.0, 2.0, 1.0]),
            covariance=np.array(
                [
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.2, 0.4],
                    [0.0, 0.2, 2.0, 0.8],
                    [0.0, 0.4, 0.8, 3.0],
                ]
            ),
        )
        self.search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    name="x", parameter_type=ParameterType.FLOAT, lower=0, upper=10
                ),
                ChoiceParameter(
                    name="z", parameter_type=ParameterType.STRING, values=["a", "b"]
                ),
            ]
        )
        self.obsf1 = ObservationFeatures({"x": 2, "z": "a"})
        self.obsf2 = ObservationFeatures({"x": 5, "z": "b"})
        self.obs1 = Observation(features=self.obsf1, data=self.obsd1)
        self.obs2 = Observation(features=self.obsf2, data=self.obsd2)
        self.t = StratifiedStandardizeY(
            search_space=self.search_space,
            observations=[self.obs1, self.obs2],
            config={"parameter_name": "z"},
        )
        self.search_space2 = SearchSpace(
            parameters=[
                RangeParameter(
                    name="x", parameter_type=ParameterType.FLOAT, lower=0, upper=10
                ),
                ChoiceParameter(
                    name="z",
                    parameter_type=ParameterType.STRING,
                    values=["a", "b", "c"],
                ),
            ]
        )
        self.strata_mapping = {"a": 0, "b": 1, "c": 1}
        self.obsf3 = ObservationFeatures({"x": 5, "z": "c"})
        self.obsd3 = ObservationData(
            metric_names=["m1", "m2", "m2"],
            means=np.array([2.0, 4.0, 16.0]),
            covariance=np.array([[1.2, 0.2, 0.4], [0.2, 2.0, 0.8], [0.4, 0.8, 3.0]]),
        )
        self.obs3 = Observation(features=self.obsf3, data=self.obsd3)
        self.t2 = StratifiedStandardizeY(
            search_space=self.search_space2,
            observations=[self.obs1, self.obs2, self.obs3],
            config={"parameter_name": "z", "strata_mapping": self.strata_mapping},
        )
        self.m1 = Metric(name="m1")
        self.m2 = Metric(name="m2")
        self.m3 = Metric(name="m3")
        self.objective = Objective(metric=self.m3, minimize=False)
        self.cons = [
            OutcomeConstraint(
                metric=self.m1, op=ComparisonOp.GEQ, bound=2.0, relative=False
            ),
            OutcomeConstraint(
                metric=self.m2, op=ComparisonOp.LEQ, bound=3.5, relative=False
            ),
        ]

    def test_Init(self) -> None:
        Ymean_expected = {
            ("m1", "a"): 1.0,
            ("m1", "b"): 3.0,
            ("m2", "a"): 5.0,
            ("m2", "b"): 1.5,
        }
        Ystd_expected = {
            ("m1", "a"): 1.0,
            ("m1", "b"): sqrt(2) * 2.0,
            ("m2", "a"): sqrt(2) * 3.0,
            ("m2", "b"): sqrt(2) * 0.5,
        }
        self.assertEqual(
            self.t.Ymean,
            Ymean_expected,
        )
        self.assertEqual(set(self.t.Ystd), set(Ystd_expected))
        for k, v in self.t.Ystd.items():
            self.assertAlmostEqual(v, Ystd_expected[k])
        with self.assertRaises(ValueError):
            # No parameter specified
            StratifiedStandardizeY(
                search_space=self.search_space,
                observations=[self.obs1, self.obs2],
            )
        with self.assertRaises(ValueError):
            # Wrong parameter type
            StratifiedStandardizeY(
                search_space=self.search_space,
                observations=[self.obs1, self.obs2],
                config={"parameter_name": "x"},
            )
        # Multiple tasks parameters
        ss3 = SearchSpace(
            parameters=[
                RangeParameter(
                    name="x", parameter_type=ParameterType.FLOAT, lower=0, upper=10
                ),
                ChoiceParameter(
                    name="z",
                    parameter_type=ParameterType.STRING,
                    values=["a", "b"],
                    is_task=True,
                    target_value="a",
                ),
                ChoiceParameter(
                    name="z2",
                    parameter_type=ParameterType.STRING,
                    values=["a", "b"],
                    is_task=True,
                    target_value="b",
                ),
            ]
        )
        with self.assertRaises(ValueError):
            StratifiedStandardizeY(
                search_space=ss3,
                observations=[self.obs1, self.obs2],
            )

        # Grab from task feature
        ss2 = SearchSpace(
            parameters=[
                RangeParameter(
                    name="x", parameter_type=ParameterType.FLOAT, lower=0, upper=10
                ),
                ChoiceParameter(
                    name="z",
                    parameter_type=ParameterType.STRING,
                    values=["a", "b"],
                    is_task=True,
                    target_value="a",
                ),
            ]
        )
        t2 = StratifiedStandardizeY(
            search_space=ss2,
            observations=[self.obs1, self.obs2],
        )
        self.assertEqual(
            t2.Ymean,
            Ymean_expected,
        )
        self.assertEqual(set(t2.Ystd), set(Ystd_expected))
        for k, v in t2.Ystd.items():
            self.assertAlmostEqual(v, Ystd_expected[k])

        # test strata_mapping
        self.assertEqual(self.t2.strata_mapping, self.strata_mapping)

    def test_TransformObservations(self) -> None:
        std_m2_a = sqrt(2) * 3
        obsd1_ta = ObservationData(
            metric_names=["m1", "m2", "m2"],
            means=np.array([0.0, -3.0 / std_m2_a, 3.0 / std_m2_a]),
            covariance=np.array(
                [
                    [1.0, 0.2 / std_m2_a, 0.4 / std_m2_a],
                    [0.2 / std_m2_a, 2.0 / 18, 0.8 / 18],
                    [0.4 / std_m2_a, 0.8 / 18, 3.0 / 18],
                ]
            ),
        )
        std_m1_b, std_m2_b = 2 * sqrt(2), sqrt(1 / 2)
        obsd1_tb = ObservationData(
            metric_names=["m1", "m2", "m2"],
            means=np.array([-2.0 / std_m1_b, 0.5 / std_m2_b, 6.5 / std_m2_b]),
            covariance=np.array(
                [
                    [1.0 / 8, 0.2 / 2, 0.4 / 2],
                    [0.2 / 2, 2.0 * 2, 0.8 * 2],
                    [0.4 / 2, 0.8 * 2, 3.0 * 2],
                ]
            ),
        )
        obsd2 = deepcopy(self.obsd1)
        obsd2 = self.t.transform_observations(
            [Observation(data=obsd2, features=ObservationFeatures({"z": "a"}))]
        )[0].data
        self.assertTrue(osd_allclose(obsd2, obsd1_ta))
        obsd2 = self.t.untransform_observations(
            [Observation(data=obsd2, features=ObservationFeatures({"z": "a"}))]
        )[0].data
        self.assertTrue(osd_allclose(obsd2, self.obsd1))
        obsd2 = deepcopy(self.obsd1)
        obsd2 = self.t.transform_observations(
            [Observation(data=obsd2, features=ObservationFeatures({"z": "b"}))]
        )[0].data
        self.assertTrue(osd_allclose(obsd2, obsd1_tb))
        obsd2 = self.t.untransform_observations(
            [Observation(data=obsd2, features=ObservationFeatures({"z": "b"}))]
        )[0].data
        self.assertTrue(osd_allclose(obsd2, self.obsd1))
        # test strata_mapping
        # "a" should be the same as above
        obsd2 = deepcopy(self.obsd1)
        obsd2 = self.t2.transform_observations(
            [Observation(data=obsd2, features=ObservationFeatures({"z": "a"}))]
        )[0].data
        self.assertTrue(osd_allclose(obsd2, obsd1_ta))
        # "b" and "c" should result in the same transformed values
        obsd2 = deepcopy(self.obsd1)
        obsd2 = self.t2.transform_observations(
            [Observation(data=obsd2, features=ObservationFeatures({"z": "b"}))]
        )[0].data
        obsd3 = deepcopy(self.obsd1)
        obsd3 = self.t2.transform_observations(
            [Observation(data=obsd3, features=ObservationFeatures({"z": "c"}))]
        )[0].data
        self.assertTrue(osd_allclose(obsd2, obsd3))
        obsd3 = self.t2.untransform_observations(
            [Observation(data=obsd3, features=ObservationFeatures({"z": "c"}))]
        )[0].data
        self.assertTrue(osd_allclose(obsd3, self.obsd1))
        obsd2 = self.t2.untransform_observations(
            [Observation(data=obsd2, features=ObservationFeatures({"z": "b"}))]
        )[0].data
        self.assertTrue(osd_allclose(obsd2, self.obsd1))

    def test_TransformOptimizationConfig(self) -> None:
        cons2 = deepcopy(self.cons)
        oc = OptimizationConfig(objective=self.objective, outcome_constraints=cons2)
        fixed_features = ObservationFeatures({"z": "a"})
        oc = self.t.transform_optimization_config(oc, None, fixed_features)
        cons_t = [
            OutcomeConstraint(
                metric=self.m1, op=ComparisonOp.GEQ, bound=1.0, relative=False
            ),
            OutcomeConstraint(
                metric=self.m2,
                op=ComparisonOp.LEQ,
                bound=(3.5 - 5.0) / (sqrt(2) * 3),
                relative=False,
            ),
        ]
        self.assertTrue(oc.outcome_constraints == cons_t)
        self.assertTrue(oc.objective == self.objective)

        # No constraints
        oc2 = OptimizationConfig(objective=self.objective)
        oc3 = deepcopy(oc2)
        fixed_features = ObservationFeatures({"z": "a"})
        oc3 = self.t.transform_optimization_config(oc3, None, fixed_features)
        self.assertTrue(oc2 == oc3)

        # Check fail with relative
        con = OutcomeConstraint(
            metric=self.m1, op=ComparisonOp.GEQ, bound=2.0, relative=True
        )
        oc = OptimizationConfig(objective=self.objective, outcome_constraints=[con])
        with self.assertRaises(ValueError):
            oc = self.t.transform_optimization_config(oc, None, fixed_features)
        # Fail without strat param fixed
        fixed_features = ObservationFeatures({"x": 2.0})
        with self.assertRaises(ValueError):
            oc = self.t.transform_optimization_config(oc, None, fixed_features)

    def test_TransformOptimizationConfigWithStrataMapping(self) -> None:
        cons2 = deepcopy(self.cons)
        oc = OptimizationConfig(objective=self.objective, outcome_constraints=cons2)
        fixed_features = ObservationFeatures({"z": "a"})
        cons_t = [
            OutcomeConstraint(
                metric=self.m1, op=ComparisonOp.GEQ, bound=1.0, relative=False
            ),
            OutcomeConstraint(
                metric=self.m2,
                op=ComparisonOp.LEQ,
                bound=(3.5 - 5.0) / (sqrt(2) * 3),
                relative=False,
            ),
        ]
        cons2 = deepcopy(self.cons)
        oc = OptimizationConfig(objective=self.objective, outcome_constraints=cons2)
        oc = self.t2.transform_optimization_config(oc, None, fixed_features)
        self.assertTrue(oc.outcome_constraints == cons_t)
        self.assertTrue(oc.objective == self.objective)
        fixed_features = ObservationFeatures({"z": "c"})
        cons2 = deepcopy(self.cons)
        oc = OptimizationConfig(objective=self.objective, outcome_constraints=cons2)
        oc = self.t2.transform_optimization_config(oc, None, fixed_features)
        cons_t2 = [
            OutcomeConstraint(
                metric=self.m1,
                op=ComparisonOp.GEQ,
                bound=(2.0 - self.t2.Ymean[("m1", 1)]) / self.t2.Ystd[("m1", 1)],
                relative=False,
            ),
            OutcomeConstraint(
                metric=self.m2,
                op=ComparisonOp.LEQ,
                bound=(3.5 - self.t2.Ymean[("m2", 1)]) / self.t2.Ystd[("m2", 1)],
                relative=False,
            ),
        ]
        self.assertEqual(oc.outcome_constraints, cons_t2)
        self.assertTrue(oc.objective == self.objective)
        cons_t = [
            OutcomeConstraint(
                metric=self.m1, op=ComparisonOp.GEQ, bound=1.0, relative=False
            ),
            OutcomeConstraint(
                metric=self.m2,
                op=ComparisonOp.LEQ,
                bound=(3.5 - 5.0) / (sqrt(2) * 3),
                relative=False,
            ),
        ]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import json
import os
import tarfile
import time
from pathlib import Path
from typing import Dict, Optional

import nbformat
import papermill
from bs4 import BeautifulSoup
from nbclient.exceptions import CellTimeoutError
from nbconvert import HTMLExporter, ScriptExporter

TUTORIALS_TO_SKIP = [
    "raytune_pytorch_cnn",  # TODO: Times out CI but passes locally. Investigate.
]


TEMPLATE = """const CWD = process.cwd();

const React = require('react');
const Tutorial = require(`${{CWD}}/core/Tutorial.js`);

class TutorialPage extends React.Component {{
  render() {{
    const {{config: siteConfig}} = this.props;
    const {{baseUrl}} = siteConfig;
    return (
      <Tutorial
        baseUrl={{baseUrl}}
        tutorialDir="{t_dir}"
        tutorialID="{tid}"
        totalExecTime={{{total_time}}}
      />
    );
  }}
}}

module.exports = TutorialPage;

"""

# we already load Plotly within html head on the site (just using <script>); this causes
# problems when trying to use requires within the notebook body, so we basically mock
# require to return already-loaded Plotly library
MOCK_JS_REQUIRES = """
<script>
const requirejs = Object();
requirejs.config = function() { };

const dependencyMap = {
    'plotly': Plotly,
}

function require(deps, fxn) {
    return fxn(...deps.map(dep => dependencyMap[dep]));
};
</script>
"""


def _get_paths(repo_dir: str, t_dir: Optional[str], tid: str) -> Dict[str, str]:
    if t_dir is not None:
        tutorial_dir = os.path.join(repo_dir, "tutorials", t_dir)
        html_dir = os.path.join(repo_dir, "website", "_tutorials", t_dir)
        js_dir = os.path.join(repo_dir, "website", "pages", "tutorials", t_dir)
        py_dir = os.path.join(repo_dir, "website", "static", "files", t_dir)

        for d in [tutorial_dir, html_dir, js_dir, py_dir]:
            os.makedirs(d, exist_ok=True)

        tutorial_path = os.path.join(tutorial_dir, "{}.ipynb".format(tid))
        html_path = os.path.join(html_dir, "{}.html".format(tid))
        js_path = os.path.join(js_dir, "{}.js".format(tid))
        ipynb_path = os.path.join(py_dir, "{}.ipynb".format(tid))
        py_path = os.path.join(py_dir, "{}.py".format(tid))
    else:
        tutorial_dir = os.path.join(repo_dir, "tutorials")
        tutorial_path = os.path.join(repo_dir, "tutorials", "{}.ipynb".format(tid))
        html_path = os.path.join(
            repo_dir, "website", "_tutorials", "{}.html".format(tid)
        )
        js_path = os.path.join(
            repo_dir, "website", "pages", "tutorials", "{}.js".format(tid)
        )
        ipynb_path = os.path.join(
            repo_dir, "website", "static", "files", "{}.ipynb".format(tid)
        )
        py_path = os.path.join(
            repo_dir, "website", "static", "files", "{}.py".format(tid)
        )

    paths = {
        "tutorial_dir": tutorial_dir,
        "tutorial_path": tutorial_path,
        "html_path": html_path,
        "js_path": js_path,
        "ipynb_path": ipynb_path,
        "py_path": py_path,
    }
    if t_dir is not None:
        paths["tar_path"] = os.path.join(py_dir, "{}.tar.gz".format(tid))
    return paths


def run_script(
    tutorial: Path, timeout_minutes: int, env: Optional[Dict[str, str]] = None
) -> None:
    if env is not None:
        os.environ.update(env)
    papermill.execute_notebook(
        tutorial,
        tutorial,
        # This timeout is on cell-execution time, not on total runtime.
        execution_timeout=timeout_minutes * 60,
    )


def gen_tutorials(
    repo_dir: str,
    exec_tutorials: bool,
    name: Optional[str] = None,
    smoke_test: bool = False,
) -> None:
    """Generate HTML tutorials for Docusaurus Ax site from Jupyter notebooks.

    Also create ipynb and py versions of tutorial in Docusaurus site for
    download.
    """
    has_errors = False

    with open(os.path.join(repo_dir, "website", "tutorials.json"), "r") as infile:
        tutorial_config = json.loads(infile.read())
    # flatten config dict
    tutorial_configs = [
        config for category in tutorial_config.values() for config in category
    ]
    # Running only the tutorial described by "name"
    if name is not None:
        tutorial_configs = [d for d in tutorial_configs if d["id"] == name]
        if len(tutorial_configs) == 0:
            raise RuntimeError(f"No tutorial found with name {name}.")
    # prepare paths for converted tutorials & files
    os.makedirs(os.path.join(repo_dir, "website", "_tutorials"), exist_ok=True)
    os.makedirs(os.path.join(repo_dir, "website", "static", "files"), exist_ok=True)
    env = {"SMOKE_TEST": "True"} if smoke_test else None

    for config in tutorial_configs:
        tid = config["id"]
        t_dir = config.get("dir")
        exec_on_build = config.get("exec_on_build", True)
        print("Generating {} tutorial".format(tid))
        paths = _get_paths(repo_dir=repo_dir, t_dir=t_dir, tid=tid)

        total_time = None

        if tid in TUTORIALS_TO_SKIP:
            print(f"Skipping execution of {tid}")
            continue
        elif exec_tutorials and exec_on_build:
            tutorial_path = Path(paths["tutorial_path"])
            print("Executing tutorial {}".format(tid))
            start_time = time.monotonic()

            # Try / catch failures for now. We will re-raise at the end.
            timeout_minutes = 15 if smoke_test else 150
            try:
                # Execute notebook.
                run_script(
                    tutorial=tutorial_path,
                    timeout_minutes=timeout_minutes,
                    env=env,
                )
                total_time = time.monotonic() - start_time
                print(
                    f"Finished executing tutorial {tid} in {total_time:.2f} seconds. "
                )
            except CellTimeoutError:
                has_errors = True
                print(
                    f"Tutorial {tid} exceeded the maximum runtime of "
                    f"{timeout_minutes} minutes."
                )
            except Exception as e:
                has_errors = True
                print(f"Encountered error running tutorial {tid}: \n {e}")

        # load notebook
        with open(paths["tutorial_path"], "r") as infile:
            nb_str = infile.read()
            nb = nbformat.reads(nb_str, nbformat.NO_CONVERT)
        # convert notebook to HTML
        exporter = HTMLExporter(template_name="classic")
        html, _ = exporter.from_notebook_node(nb)

        # pull out html div for notebook
        soup = BeautifulSoup(html, "html.parser")
        nb_meat = soup.find("div", {"id": "notebook-container"})
        del nb_meat.attrs["id"]
        nb_meat.attrs["class"] = ["notebook"]

        # when output html, iframe it (useful for Ax reports)
        for html_div in nb_meat.findAll("div", {"class": "output_html"}):
            if html_div.html is not None:
                iframe = soup.new_tag("iframe")
                iframe.attrs["src"] = "data:text/html;charset=utf-8," + str(
                    html_div.html
                )
                # replace `#` in CSS
                iframe.attrs["src"] = iframe.attrs["src"].replace("#", "%23")
                html_div.contents = [iframe]

        html_out = MOCK_JS_REQUIRES + str(nb_meat)

        # generate HTML file
        with open(paths["html_path"], "w") as html_outfile:
            html_outfile.write(html_out)

        # generate JS file
        t_dir_js = t_dir if t_dir else ""
        script = TEMPLATE.format(
            t_dir=t_dir_js,
            tid=tid,
            total_time=total_time if total_time is not None else "null",
        )
        with open(paths["js_path"], "w") as js_outfile:
            js_outfile.write(script)

        # output tutorial in both ipynb & py form
        nbformat.write(nb, paths["ipynb_path"])
        exporter = ScriptExporter()
        script, _ = exporter.from_notebook_node(nb)
        with open(paths["py_path"], "w") as py_outfile:
            py_outfile.write(script)

        # create .tar archive (if necessary)
        if t_dir is not None:
            with tarfile.open(paths["tar_path"], "w:gz") as tar:
                tar.add(
                    paths["tutorial_dir"],
                    arcname=os.path.basename(paths["tutorial_dir"]),
                )

    if has_errors:
        raise Exception("There are errors in tutorials, will not continue to publish")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Generate JS, HTML, ipynb, and py files for tutorials."
    )
    parser.add_argument(
        "-w", "--repo_dir", metavar="path", required=True, help="Ax repo directory."
    )
    parser.add_argument(
        "-s", "--smoke", action="store_true", help="Run in smoke test mode."
    )
    parser.add_argument(
        "-e",
        "--exec_tutorials",
        action="store_true",
        default=False,
        help="Execute tutorials (instead of just converting).",
    )
    parser.add_argument(
        "-n",
        "--name",
        help="Run a specific tutorial by name. The name should not include the "
        ".ipynb extension.",
    )
    args = parser.parse_args()
    gen_tutorials(
        args.repo_dir,
        args.exec_tutorials,
        smoke_test=args.smoke,
        name=args.name,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from copy import deepcopy
from typing import Dict, List, Union

from ax.utils.common.base import SortableBase
from ax.utils.common.equality import equality_typechecker


class RiskMeasure(SortableBase):
    """A class for defining risk measures.

    This can be used with a `RobustSearchSpace`, to convert the predictions over
    `ParameterDistribution`s to robust metrics, which then get used in candidate
    generation to recommend robust candidates.

    See `ax/modelbridge/modelbridge_utils.py` for `RISK_MEASURE_NAME_TO_CLASS`,
    which lists the supported risk measures, and for `extract_risk_measure`
    helper, which extracts the BoTorch risk measure.
    """

    def __init__(
        self,
        risk_measure: str,
        options: Dict[str, Union[int, float, bool, List[float]]],
    ) -> None:
        """Initialize a risk measure.

        Args:
            risk_measure: The name of the risk measure to use. This should have a
                corresponding entry in `RISK_MEASURE_NAME_TO_CLASS`.
            options: A dictionary of keyword arguments for initializing the risk
                measure. Except for MARS, the risk measure will be initialized as
                `RISK_MEASURE_NAME_TO_CLASS[risk_measure](**options)`. For MARS,
                additional attributes are needed to inform the scalarization.
        """
        super().__init__()
        self.risk_measure = risk_measure
        self.options = options

    def clone(self) -> RiskMeasure:
        """Clone."""
        return RiskMeasure(
            risk_measure=self.risk_measure,
            options=deepcopy(self.options),
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            "risk_measure=" + self.risk_measure + ", "
            "options=" + repr(self.options) + ")"
        )

    @property
    def _unique_id(self) -> str:
        return str(self)

    @equality_typechecker
    def __eq__(self, other: RiskMeasure) -> bool:
        return str(self) == str(other)



#!/usr/bin/env fbpython
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.exceptions.core import AxWarning
from ax.telemetry.generation_strategy import GenerationStrategyCreatedRecord
from ax.utils.common.testutils import TestCase
from ax.utils.testing.modeling_stubs import (
    get_generation_strategy,
    sobol_gpei_generation_node_gs,
)


class TestGenerationStrategy(TestCase):
    def test_generation_strategy_created_record_from_generation_strategy(self) -> None:
        gs = get_generation_strategy()
        record = GenerationStrategyCreatedRecord.from_generation_strategy(
            generation_strategy=gs
        )
        expected = GenerationStrategyCreatedRecord(
            generation_strategy_name="Sobol+BO_MIXED",
            num_requested_initialization_trials=6,
            num_requested_bayesopt_trials=-1,
            num_requested_other_trials=0,
            max_parallelism=3,
        )
        self.assertEqual(record, expected)

    def test_generation_strategy_created_record_node_based(self) -> None:
        gs = sobol_gpei_generation_node_gs()
        with self.assertWarnsRegex(
            AxWarning,
            "`GenerationStrategyCreatedRecord` does not fully support node-based*",
        ):
            record = GenerationStrategyCreatedRecord.from_generation_strategy(
                generation_strategy=gs
            )
        expected = GenerationStrategyCreatedRecord(
            generation_strategy_name="Sobol+MBM_Nodes",
            num_requested_initialization_trials=None,
            num_requested_bayesopt_trials=None,
            num_requested_other_trials=None,
            max_parallelism=None,
        )
        self.assertEqual(record, expected)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest.mock import patch

import numpy as np
from ax.core.arm import Arm
from ax.core.data import Data
from ax.core.generator_run import GeneratorRun
from ax.core.metric import Metric
from ax.core.objective import MultiObjective, Objective
from ax.core.observation import ObservationData, ObservationFeatures
from ax.core.optimization_config import MultiObjectiveOptimizationConfig
from ax.core.outcome_constraint import (
    ObjectiveThreshold,
    OutcomeConstraint,
    ScalarizedOutcomeConstraint,
)
from ax.core.types import ComparisonOp
from ax.core.utils import get_pending_observation_features
from ax.modelbridge.modelbridge_utils import (
    extract_objective_thresholds,
    extract_outcome_constraints,
    feasible_hypervolume,
    observation_data_to_array,
    pending_observations_as_array_list,
)
from ax.modelbridge.registry import Models
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import not_none
from ax.utils.testing.core_stubs import (
    get_experiment,
    get_hierarchical_search_space_experiment,
)
from pyre_extensions import none_throws


TEST_PARAMETERIZATON_LIST = ["5", "foo", "True", "5"]


class TestModelbridgeUtils(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.experiment = get_experiment()
        self.arm = Arm({"x": 5, "y": "foo", "z": True, "w": 5})
        self.trial = self.experiment.new_trial(GeneratorRun([self.arm]))
        self.experiment_2 = get_experiment()
        self.batch_trial = self.experiment_2.new_batch_trial(GeneratorRun([self.arm]))
        self.batch_trial.set_status_quo_with_weight(self.experiment_2.status_quo, 1)
        self.obs_feat = ObservationFeatures.from_arm(
            arm=self.trial.arm, trial_index=self.trial.index
        )
        self.hss_exp = get_hierarchical_search_space_experiment()
        self.hss_sobol = Models.SOBOL(search_space=self.hss_exp.search_space)
        self.hss_gr = self.hss_sobol.gen(n=1)
        self.hss_trial = self.hss_exp.new_trial(self.hss_gr)
        self.hss_arm = not_none(self.hss_trial.arm)
        self.hss_cand_metadata = self.hss_trial._get_candidate_metadata(
            arm_name=self.hss_arm.name
        )
        self.hss_full_parameterization = self.hss_cand_metadata.get(
            Keys.FULL_PARAMETERIZATION
        ).copy()
        self.assertTrue(
            all(
                p_name in self.hss_full_parameterization
                for p_name in self.hss_exp.search_space.parameters
            )
        )
        self.hss_obs_feat = ObservationFeatures.from_arm(
            arm=self.hss_arm,
            trial_index=self.hss_trial.index,
            metadata=self.hss_cand_metadata,
        )
        self.hss_obs_feat_all_params = ObservationFeatures.from_arm(
            arm=Arm(self.hss_full_parameterization),
            trial_index=self.hss_trial.index,
            metadata={Keys.FULL_PARAMETERIZATION: self.hss_full_parameterization},
        )

    def test_extract_outcome_constraints(self) -> None:
        outcomes = ["m1", "m2", "m3"]
        # pass no outcome constraints
        self.assertIsNone(extract_outcome_constraints([], outcomes))

        outcome_constraints = [
            OutcomeConstraint(metric=Metric("m1"), op=ComparisonOp.LEQ, bound=0)
        ]
        res = extract_outcome_constraints(outcome_constraints, outcomes)
        # pyre-fixme[16]: Optional type has no attribute `__getitem__`.
        self.assertEqual(res[0].shape, (1, 3))
        self.assertListEqual(list(res[0][0]), [1, 0, 0])
        self.assertEqual(res[1][0][0], 0)

        outcome_constraints = [
            OutcomeConstraint(metric=Metric("m1"), op=ComparisonOp.LEQ, bound=0),
            ScalarizedOutcomeConstraint(
                metrics=[Metric("m2"), Metric("m3")],
                weights=[0.5, 0.5],
                op=ComparisonOp.GEQ,
                bound=1,
            ),
        ]
        res = extract_outcome_constraints(outcome_constraints, outcomes)
        self.assertEqual(res[0].shape, (2, 3))
        self.assertListEqual(list(res[0][0]), [1, 0, 0])
        self.assertListEqual(list(res[0][1]), [0, -0.5, -0.5])
        self.assertEqual(res[1][0][0], 0)
        self.assertEqual(res[1][1][0], -1)

    def test_extract_objective_thresholds(self) -> None:
        outcomes = ["m1", "m2", "m3", "m4"]
        objective = MultiObjective(
            objectives=[
                Objective(metric=Metric(name), minimize=False) for name in outcomes[:3]
            ]
        )
        objective_thresholds = [
            ObjectiveThreshold(
                metric=Metric(name),
                op=ComparisonOp.LEQ,
                bound=float(i + 2),
                relative=False,
            )
            for i, name in enumerate(outcomes[:3])
        ]

        # None of no thresholds
        self.assertIsNone(
            extract_objective_thresholds(
                objective_thresholds=[], objective=objective, outcomes=outcomes
            )
        )

        # Working case
        obj_t = extract_objective_thresholds(
            objective_thresholds=objective_thresholds,
            objective=objective,
            outcomes=outcomes,
        )
        expected_obj_t_not_nan = np.array([2.0, 3.0, 4.0])
        # pyre-fixme[16]: Optional type has no attribute `__getitem__`.
        self.assertTrue(np.array_equal(obj_t[:3], expected_obj_t_not_nan[:3]))
        self.assertTrue(np.isnan(obj_t[-1]))
        # pyre-fixme[16]: Optional type has no attribute `shape`.
        self.assertEqual(obj_t.shape[0], 4)

        # Returns NaN for objectives without a threshold.
        obj_t = extract_objective_thresholds(
            objective_thresholds=objective_thresholds[:2],
            objective=objective,
            outcomes=outcomes,
        )
        self.assertTrue(np.array_equal(obj_t[:2], expected_obj_t_not_nan[:2]))
        self.assertTrue(np.isnan(obj_t[-2:]).all())

        # Fails if a threshold does not have a corresponding metric.
        objective2 = Objective(Metric("m1"), minimize=False)
        with self.assertRaisesRegex(ValueError, "corresponding metrics"):
            extract_objective_thresholds(
                objective_thresholds=objective_thresholds,
                objective=objective2,
                outcomes=outcomes,
            )

        # Works with a single objective, single threshold
        obj_t = extract_objective_thresholds(
            objective_thresholds=objective_thresholds[:1],
            objective=objective2,
            outcomes=outcomes,
        )
        self.assertEqual(obj_t[0], 2.0)
        self.assertTrue(np.all(np.isnan(obj_t[1:])))
        self.assertEqual(obj_t.shape[0], 4)

        # Fails if relative
        objective_thresholds[2] = ObjectiveThreshold(
            metric=Metric("m3"), op=ComparisonOp.LEQ, bound=3
        )
        with self.assertRaises(ValueError):
            extract_objective_thresholds(
                objective_thresholds=objective_thresholds,
                objective=objective,
                outcomes=outcomes,
            )
        objective_thresholds[2] = ObjectiveThreshold(
            metric=Metric("m3"), op=ComparisonOp.LEQ, bound=3, relative=True
        )
        with self.assertRaises(ValueError):
            extract_objective_thresholds(
                objective_thresholds=objective_thresholds,
                objective=objective,
                outcomes=outcomes,
            )

    def test_observation_data_to_array(self) -> None:
        outcomes = ["a", "b", "c"]
        obsd = ObservationData(
            metric_names=["c", "a", "b"],
            means=np.array([1, 2, 3]),
            covariance=np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
        )
        Y, Ycov = observation_data_to_array(outcomes=outcomes, observation_data=[obsd])
        self.assertTrue(np.array_equal(Y, np.array([[2, 3, 1]])))
        self.assertTrue(
            np.array_equal(Ycov, np.array([[[5, 6, 4], [8, 9, 7], [2, 3, 1]]]))
        )

        # With missing metrics.
        obsd2 = ObservationData(
            metric_names=["c", "a"],
            means=np.array([1, 2]),
            covariance=np.array([[1, 2], [4, 5]]),
        )
        Y, Ycov = observation_data_to_array(
            outcomes=outcomes, observation_data=[obsd, obsd2]
        )
        nan = float("nan")
        self.assertTrue(
            np.array_equal(Y, np.array([[2, 3, 1], [2, nan, 1]]), equal_nan=True)
        )
        self.assertTrue(
            np.array_equal(
                Ycov,
                np.array(
                    [
                        [[5, 6, 4], [8, 9, 7], [2, 3, 1]],
                        [[5, nan, 4], [nan, nan, nan], [2, nan, 1]],
                    ]
                ),
                equal_nan=True,
            )
        )

    def test_feasible_hypervolume(self) -> None:
        ma = Metric(name="a", lower_is_better=False)
        mb = Metric(name="b", lower_is_better=True)
        mc = Metric(name="c", lower_is_better=False)
        optimization_config = MultiObjectiveOptimizationConfig(
            objective=MultiObjective(metrics=[ma, mb]),
            outcome_constraints=[
                OutcomeConstraint(
                    mc,
                    op=ComparisonOp.GEQ,
                    bound=0,
                    relative=False,
                )
            ],
            objective_thresholds=[
                ObjectiveThreshold(
                    ma,
                    bound=1.0,
                ),
                ObjectiveThreshold(
                    mb,
                    bound=1.0,
                ),
            ],
        )
        feas_hv = feasible_hypervolume(
            optimization_config,
            values={
                "a": np.array(
                    [
                        1.0,
                        3.0,
                        2.0,
                        2.0,
                    ]
                ),
                "b": np.array(
                    [
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                    ]
                ),
                "c": np.array(
                    [
                        0.0,
                        -0.0,
                        1.0,
                        -2.0,
                    ]
                ),
            },
        )
        self.assertEqual(list(feas_hv), [0.0, 0.0, 1.0, 1.0])

    def test_pending_observations_as_array_list(self) -> None:
        # Mark a trial dispatched so that there are pending observations.
        self.trial.mark_running(no_runner_required=True)
        # If outcome names are respected, unlisted metrics should be filtered out.
        self.assertEqual(
            [
                x.tolist()
                for x in none_throws(
                    pending_observations_as_array_list(
                        pending_observations=none_throws(
                            get_pending_observation_features(self.experiment)
                        ),
                        outcome_names=["m2", "m1"],
                        param_names=["x", "y", "z", "w"],
                    )
                )
            ],
            [[TEST_PARAMETERIZATON_LIST], [TEST_PARAMETERIZATON_LIST]],
        )
        self.experiment.attach_data(
            Data.from_evaluations(
                {self.trial.arm.name: {"m2": (1, 0)}}, trial_index=self.trial.index
            )
        )
        # With `fetch_data` on trial returning data for metric "m2", that metric
        # should no longer have pending observation features.
        with patch.object(
            self.trial,
            "fetch_data",
            return_value=Data.from_evaluations(
                {self.trial.arm.name: {"m2": (1, 0)}}, trial_index=self.trial.index
            ),
        ):
            pending = none_throws(get_pending_observation_features(self.experiment))
        # There should be no pending observations for metric m2 now, since the
        # only trial there is, has been updated with data for it.
        self.assertEqual(
            [
                x.tolist()
                for x in none_throws(
                    pending_observations_as_array_list(
                        pending_observations=pending,
                        outcome_names=["m2", "m1"],
                        param_names=["x", "y", "z", "w"],
                    )
                )
            ],
            [[], [TEST_PARAMETERIZATON_LIST]],
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from typing import Type

from ax.core.map_data import MapData, MapKeyInfo
from ax.core.metric import Metric, MetricFetchE
from ax.utils.common.result import Result

MapMetricFetchResult = Result[MapData, MetricFetchE]


class MapMetric(Metric):
    """Base class for representing metrics that return `MapData`.

    The `fetch_trial_data` method is the essential method to override when
    subclassing, which specifies how to retrieve a Metric, for a given trial.

    A MapMetric must return a MapData object, which requires (at minimum) the following:
        https://ax.dev/api/_modules/ax/core/data.html#Data.required_columns

    Attributes:
        lower_is_better: Flag for metrics which should be minimized.
        properties: Properties specific to a particular metric.
    """

    data_constructor: Type[MapData] = MapData
    map_key_info: MapKeyInfo[float] = MapKeyInfo(key="steps", default_value=0.0)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

# necessary to import this file so SQLAlchemy knows about the event listeners
# see https://fburl.com/8mn7yjt2
from ax.storage.sqa_store import validation
from ax.storage.sqa_store.load import load_experiment as sqa_load
from ax.storage.sqa_store.save import save_experiment as sqa_save


__all__ = ["sqa_load", "sqa_save"]

del validation



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from ax.modelbridge.transforms.unit_x import UnitX


class CenteredUnitX(UnitX):
    """Map X to [-1, 1]^d for RangeParameter of type float and not log scale.

    Transform is done in-place.
    """

    target_lb: float = -1.0
    target_range: float = 2.0



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from abc import ABC, abstractmethod
from typing import Any, Callable, TypeVar

T = TypeVar("T")


class ClassDecorator(ABC):
    """
    Template for making a decorator work as a class level decorator.  That decorator
    should extend `ClassDecorator`.  It must implement `__init__` and
    `decorate_callable`.  See `disable_logger.decorate_callable` for an example.
    `decorate_callable` should call `self._call_func()` instead of directly calling
    `func` to handle static functions.
    Note: `_call_func` is still imperfect and unit tests should be used to ensure
    everything is working properly.  There is a lot of complexity in detecting
    classmethods and staticmethods and removing the self argument in the right
    situations. For best results always use keyword args in the decorated class.

    `DECORATE_PRIVATE` can be set to determine whether private methods should be
    decorated. In the case of a logging decorator, you may only want to decorate things
    the user calls. But in the case of a disable logging decorator, you may want to
    decorate everything to ensure no logs escape.
    """

    DECORATE_PRIVATE = True

    def decorate_class(self, klass: T) -> T:
        for attr in dir(klass):
            if not self.DECORATE_PRIVATE and attr[0] == "_":
                continue

            attr_value = getattr(klass, attr)
            if (
                not callable(attr_value)
                or isinstance(attr_value, type)
                or attr
                in (
                    "__subclasshook__",
                    "__class__",
                    "__repr__",
                    "__str__",
                    "__getattribute__",
                    "__new__",
                    "__call__",
                    "__eq__",
                    "_call_func",
                )
            ):
                continue

            setattr(klass, attr, self.decorate_callable(attr_value))
        return klass

    @abstractmethod
    def decorate_callable(self, func: Callable[..., T]) -> Callable[..., T]:
        pass

    def __call__(self, func: Callable[..., T]) -> Callable[..., T]:
        if isinstance(func, type):
            return self.decorate_class(func)
        return self.decorate_callable(func)

    def _call_func(self, func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
        try:
            return func(*args, **kwargs)
        except TypeError as e:
            # static functions
            try:
                return func(*args[1:], **kwargs)
            except TypeError:
                # it wasn't that it was a static function
                raise e



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from dataclasses import dataclass
from typing import List, Union

import numpy as np
import torch
from ax.core.metric import Metric
from ax.core.objective import MultiObjective
from ax.core.optimization_config import MultiObjectiveOptimizationConfig
from ax.core.outcome_constraint import ObjectiveThreshold, OutcomeConstraint
from ax.core.parameter import ChoiceParameter, ParameterType, RangeParameter
from ax.core.risk_measures import RiskMeasure
from ax.core.search_space import RobustSearchSpace, SearchSpace
from ax.core.types import ComparisonOp
from ax.exceptions.core import UserInputError
from ax.modelbridge.modelbridge_utils import (
    _array_to_tensor,
    extract_risk_measure,
    extract_robust_digest,
    feasible_hypervolume,
    process_contextual_datasets,
    RISK_MEASURE_NAME_TO_CLASS,
    transform_search_space,
)
from ax.modelbridge.registry import Cont_X_trans, Y_trans
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import not_none
from ax.utils.testing.core_stubs import get_robust_search_space, get_search_space
from botorch.acquisition.risk_measures import VaR
from botorch.utils.datasets import ContextualDataset, SupervisedDataset


class TestModelBridgeUtils(TestCase):
    def test__array_to_tensor(self) -> None:
        from ax.modelbridge import ModelBridge

        @dataclass
        class MockModelbridge(ModelBridge):
            def _array_to_tensor(self, array: Union[np.ndarray, List[float]]):
                return _array_to_tensor(array=array)

        mock_modelbridge = MockModelbridge()
        arr = [0.0]
        res = _array_to_tensor(array=arr)
        self.assertEqual(len(res.size()), 1)
        self.assertEqual(res.size()[0], 1)

        res = _array_to_tensor(array=arr, modelbridge=mock_modelbridge)
        self.assertEqual(len(res.size()), 1)
        self.assertEqual(res.size()[0], 1)

    def test_extract_risk_measure(self) -> None:
        rm = RiskMeasure(
            risk_measure="VaR",
            options={"alpha": 0.8, "n_w": 5},
        )
        rm_module = extract_risk_measure(risk_measure=rm)
        self.assertIsInstance(rm_module, VaR)
        self.assertEqual(rm_module.alpha, 0.8)
        self.assertEqual(rm_module.n_w, 5)

        # Test unknown risk measure.
        with self.assertRaisesRegex(UserInputError, "constructing"):
            extract_risk_measure(
                risk_measure=RiskMeasure(
                    risk_measure="VVar",
                    options={},
                )
            )
        # Test invalid options.
        with self.assertRaisesRegex(UserInputError, "constructing"):
            extract_risk_measure(
                risk_measure=RiskMeasure(
                    risk_measure="VaR",
                    options={"alpha": 5, "n_w": 5},
                )
            )

        # Test using user-defined risk measures.

        class CustomRM(VaR):
            pass

        RISK_MEASURE_NAME_TO_CLASS["custom"] = CustomRM

        rm = RiskMeasure(
            risk_measure="custom",
            options={"alpha": 0.8, "n_w": 5},
        )
        self.assertEqual(rm.risk_measure, "custom")
        self.assertIsInstance(extract_risk_measure(risk_measure=rm), CustomRM)

    def test_extract_robust_digest(self) -> None:
        # Test with non-robust search space.
        ss = get_search_space()
        self.assertIsNone(extract_robust_digest(ss, list(ss.parameters)))
        # Test with non-environmental search space.
        for multiplicative in (True, False):
            rss = get_robust_search_space(num_samples=8)
            if multiplicative:
                for p in rss.parameter_distributions:
                    p.multiplicative = True
                rss.multiplicative = True
            robust_digest = not_none(extract_robust_digest(rss, list(rss.parameters)))
            self.assertEqual(robust_digest.multiplicative, multiplicative)
            self.assertEqual(robust_digest.environmental_variables, [])
            self.assertIsNone(robust_digest.sample_environmental)
            samples = not_none(robust_digest.sample_param_perturbations)()
            self.assertEqual(samples.shape, (8, 4))
            constructor = np.ones if multiplicative else np.zeros
            self.assertTrue(np.equal(samples[:, 2:], constructor((8, 2))).all())
            # Exponential distribution is non-negative, so we can check for that.
            self.assertTrue(np.all(samples[:, 1] > 0))
            # Check that it works as expected if param_names is missing some
            # non-distributional parameters.
            robust_digest = not_none(
                extract_robust_digest(rss, list(rss.parameters)[:-1])
            )
            samples = not_none(robust_digest.sample_param_perturbations)()
            self.assertEqual(samples.shape, (8, 3))
            self.assertTrue(np.equal(samples[:, 2:], constructor((8, 1))).all())
            self.assertTrue(np.all(samples[:, 1] > 0))
            # Check that it errors if we're missing distributional parameters.
            with self.assertRaisesRegex(RuntimeError, "All distributional"):
                extract_robust_digest(rss, list(rss.parameters)[1:])
        # Test with environmental search space.
        all_params = list(rss.parameters.values())
        rss = RobustSearchSpace(
            parameters=all_params[2:],
            parameter_distributions=rss.parameter_distributions,
            num_samples=8,
            environmental_variables=all_params[:2],
        )
        robust_digest = not_none(extract_robust_digest(rss, list(rss.parameters)))
        self.assertFalse(robust_digest.multiplicative)
        self.assertIsNone(robust_digest.sample_param_perturbations)
        self.assertEqual(robust_digest.environmental_variables, ["x", "y"])
        samples = not_none(robust_digest.sample_environmental)()
        self.assertEqual(samples.shape, (8, 2))
        # Both are continuous distributions, should be non-zero.
        self.assertTrue(np.all(samples != 0))
        # Check for error if environmental variables are not at the end.
        with self.assertRaisesRegex(RuntimeError, "last entries"):
            extract_robust_digest(rss, list(rss.parameters)[::-1])
        # Test with mixed search space.
        rss = RobustSearchSpace(
            parameters=all_params[1:],
            parameter_distributions=rss.parameter_distributions,
            num_samples=8,
            environmental_variables=all_params[:1],
        )
        robust_digest = not_none(extract_robust_digest(rss, list(rss.parameters)))
        self.assertFalse(robust_digest.multiplicative)
        self.assertEqual(
            not_none(robust_digest.sample_param_perturbations)().shape, (8, 3)
        )
        self.assertEqual(not_none(robust_digest.sample_environmental)().shape, (8, 1))
        self.assertEqual(robust_digest.environmental_variables, ["x"])

    def test_feasible_hypervolume(self) -> None:
        ma = Metric(name="a", lower_is_better=False)
        mb = Metric(name="b", lower_is_better=True)
        mc = Metric(name="c", lower_is_better=False)
        optimization_config = MultiObjectiveOptimizationConfig(
            objective=MultiObjective(metrics=[ma, mb]),
            outcome_constraints=[
                OutcomeConstraint(
                    mc,
                    op=ComparisonOp.GEQ,
                    bound=0,
                    relative=False,
                )
            ],
            objective_thresholds=[
                ObjectiveThreshold(
                    ma,
                    bound=1.0,
                ),
                ObjectiveThreshold(
                    mb,
                    bound=1.0,
                ),
            ],
        )
        feas_hv = feasible_hypervolume(
            optimization_config,
            values={
                "a": np.array(
                    [
                        1.0,
                        3.0,
                        2.0,
                        2.0,
                    ]
                ),
                "b": np.array(
                    [
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                    ]
                ),
                "c": np.array(
                    [
                        0.0,
                        -0.0,
                        1.0,
                        -2.0,
                    ]
                ),
            },
        )
        self.assertEqual(list(feas_hv), [0.0, 0.0, 1.0, 1.0])

    def test_get_transformed_dimensionality(self) -> None:
        search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    name="range",
                    parameter_type=ParameterType.FLOAT,
                    lower=1,
                    upper=8,
                ),
                ChoiceParameter(
                    name="choice",
                    parameter_type=ParameterType.INT,
                    values=[11, 18, 1998],
                    is_ordered=False,
                ),
            ]
        )

        transformed_search_space = transform_search_space(
            search_space=search_space,
            transforms=Cont_X_trans + Y_trans,
            transform_configs={},
        )

        expected = SearchSpace(
            parameters=[
                RangeParameter(
                    name="range", parameter_type=ParameterType.FLOAT, lower=0, upper=1
                ),
                RangeParameter(
                    name="choice_OH_PARAM__0",
                    parameter_type=ParameterType.FLOAT,
                    lower=0,
                    upper=1,
                ),
                RangeParameter(
                    name="choice_OH_PARAM__1",
                    parameter_type=ParameterType.FLOAT,
                    lower=0,
                    upper=1,
                ),
                RangeParameter(
                    name="choice_OH_PARAM__2",
                    parameter_type=ParameterType.FLOAT,
                    lower=0,
                    upper=1,
                ),
            ],
            parameter_constraints=[],
        )

        self.assertEqual(transformed_search_space, expected)

    def test_process_contextual_datasets(self) -> None:
        num_samples = 5
        num_contexts = 3
        feature_names = [f"x_c{i}" for i in range(num_contexts)]
        parameter_decomposition = {
            f"context_{i}": [f"x_c{i}"] for i in range(num_contexts)
        }
        context_buckets = list(parameter_decomposition.keys())
        context_outcome_list = [f"y:context_{i}" for i in range(num_contexts)]
        metric_decomposition = {f"{c}": [f"y:{c}"] for c in context_buckets}

        X = torch.rand(num_samples, num_contexts)

        dataset_list = [
            SupervisedDataset(
                X=X,
                Y=torch.rand(num_samples, 1),
                Yvar=torch.rand(num_samples, 1),
                feature_names=feature_names,
                outcome_names=["m1_overall"],
            ),
            SupervisedDataset(
                X=X,
                Y=torch.rand(num_samples, 1),
                Yvar=torch.rand(num_samples, 1),
                feature_names=feature_names,
                outcome_names=["m2_overall"],
            ),
        ]
        # process dataset list with overall outcome only
        contextual_datasets = process_contextual_datasets(
            datasets=dataset_list,
            outcomes=["m1_overall", "m2_overall"],
            parameter_decomposition=parameter_decomposition,
        )
        self.assertEqual(len(contextual_datasets), 2)
        for d in contextual_datasets:
            self.assertIsInstance(d, ContextualDataset)
            self.assertEqual(len(d.outcome_names), 1)

        for m in context_outcome_list:
            dataset_list.append(
                SupervisedDataset(
                    X=X,
                    Y=torch.rand(num_samples, 1),
                    Yvar=torch.rand(num_samples, 1),
                    feature_names=feature_names,
                    outcome_names=[m],
                )
            )
        # # process dataset list with context-level outcomes
        contextual_datasets = process_contextual_datasets(
            datasets=dataset_list[2:],
            outcomes=context_outcome_list,
            parameter_decomposition=parameter_decomposition,
            metric_decomposition=metric_decomposition,
        )
        self.assertEqual(len(contextual_datasets), 1)
        self.assertIsInstance(contextual_datasets[0], ContextualDataset)
        self.assertListEqual(contextual_datasets[0].outcome_names, context_outcome_list)

        # process dataset list with overall outcome and context-level outcomes
        contextual_datasets = process_contextual_datasets(
            datasets=dataset_list,
            outcomes=["m1_overall", "m2_overall"] + context_outcome_list,
            parameter_decomposition=parameter_decomposition,
            metric_decomposition=metric_decomposition,
        )
        self.assertEqual(len(contextual_datasets), 3)
        for d in contextual_datasets:
            self.assertIsInstance(d, ContextualDataset)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.storage.sqa_store.decoder import Decoder
from ax.storage.sqa_store.encoder import Encoder
from ax.utils.testing.core_stubs import (
    get_abandoned_arm,
    get_arm,
    get_batch_trial,
    get_branin_metric,
    get_branin_objective,
    get_branin_outcome_constraint,
    get_choice_parameter,
    get_experiment_with_batch_and_single_trial,
    get_experiment_with_batch_trial,
    get_experiment_with_data,
    get_experiment_with_map_data,
    get_experiment_with_multi_objective,
    get_experiment_with_scalarized_objective_and_outcome_constraint,
    get_factorial_metric,
    get_fixed_parameter,
    get_generator_run,
    get_generator_run2,
    get_hartmann_metric,
    get_hierarchical_search_space_experiment,
    get_metric,
    get_objective,
    get_order_constraint,
    get_ordered_choice_parameter,
    get_outcome_constraint,
    get_parameter_constraint,
    get_range_parameter,
    get_risk_measure,
    get_robust_search_space,
    get_robust_search_space_environmental,
    get_scalarized_objective,
    get_scalarized_outcome_constraint,
    get_sum_constraint1,
    get_sum_constraint2,
    get_synthetic_runner,
    get_task_choice_parameter,
    get_trial,
)


# pyre-fixme[5]: Global expression must be annotated.
TEST_CASES = [
    (
        "AbandonedArm",
        get_abandoned_arm,
        Encoder.abandoned_arm_to_sqa,
        Decoder.abandoned_arm_from_sqa,
    ),
    ("Arm", get_arm, Encoder.arm_to_sqa, Decoder.arm_from_sqa),
    ("BatchTrial", get_batch_trial, Encoder.trial_to_sqa, Decoder.trial_from_sqa),
    ("BraninMetric", get_branin_metric, Encoder.metric_to_sqa, Decoder.metric_from_sqa),
    (
        "BraninObjective",
        get_branin_objective,
        Encoder.objective_to_sqa,
        Decoder.metric_from_sqa,
    ),
    (
        "BraninOutcomeConstraint",
        get_branin_outcome_constraint,
        Encoder.outcome_constraint_to_sqa,
        Decoder.metric_from_sqa,
    ),
    (
        "ChoiceParameter",
        get_choice_parameter,
        Encoder.parameter_to_sqa,
        Decoder.parameter_from_sqa,
    ),
    (
        "ChoiceParameter",
        get_ordered_choice_parameter,
        Encoder.parameter_to_sqa,
        Decoder.parameter_from_sqa,
    ),
    (
        "ChoiceParameter",
        get_task_choice_parameter,
        Encoder.parameter_to_sqa,
        Decoder.parameter_from_sqa,
    ),
    (
        "Experiment",
        get_experiment_with_batch_trial,
        Encoder.experiment_to_sqa,
        Decoder.experiment_from_sqa,
    ),
    (
        "Experiment",
        get_experiment_with_batch_and_single_trial,
        Encoder.experiment_to_sqa,
        Decoder.experiment_from_sqa,
    ),
    (
        "Experiment",
        get_experiment_with_data,
        Encoder.experiment_to_sqa,
        Decoder.experiment_from_sqa,
    ),
    (
        "Experiment",
        get_experiment_with_multi_objective,
        Encoder.experiment_to_sqa,
        Decoder.experiment_from_sqa,
    ),
    (
        "Experiment",
        get_experiment_with_scalarized_objective_and_outcome_constraint,
        Encoder.experiment_to_sqa,
        Decoder.experiment_from_sqa,
    ),
    (
        "Experiment",
        get_experiment_with_map_data,
        Encoder.experiment_to_sqa,
        Decoder.experiment_from_sqa,
    ),
    (
        "Experiment",
        get_hierarchical_search_space_experiment,
        Encoder.experiment_to_sqa,
        Decoder.experiment_from_sqa,
    ),
    (
        "FixedParameter",
        get_fixed_parameter,
        Encoder.parameter_to_sqa,
        Decoder.parameter_from_sqa,
    ),
    (
        "FactorialMetric",
        get_factorial_metric,
        Encoder.metric_to_sqa,
        Decoder.metric_from_sqa,
    ),
    (
        "GeneratorRun",
        get_generator_run,
        Encoder.generator_run_to_sqa,
        Decoder.generator_run_from_sqa,
    ),
    (
        "GeneratorRun",
        get_generator_run2,
        Encoder.generator_run_to_sqa,
        Decoder.generator_run_from_sqa,
    ),
    (
        "HartmannMetric",
        get_hartmann_metric,
        Encoder.metric_to_sqa,
        Decoder.metric_from_sqa,
    ),
    ("Metric", get_metric, Encoder.metric_to_sqa, Decoder.metric_from_sqa),
    ("Objective", get_objective, Encoder.objective_to_sqa, Decoder.metric_from_sqa),
    (
        "OrderConstraint",
        get_order_constraint,
        Encoder.parameter_constraint_to_sqa,
        Decoder.parameter_constraint_from_sqa,
    ),
    (
        "OutcomeConstraint",
        get_outcome_constraint,
        Encoder.outcome_constraint_to_sqa,
        Decoder.metric_from_sqa,
    ),
    (
        "ParameterConstraint",
        get_parameter_constraint,
        Encoder.parameter_constraint_to_sqa,
        Decoder.parameter_constraint_from_sqa,
    ),
    (
        "RangeParameter",
        get_range_parameter,
        Encoder.parameter_to_sqa,
        Decoder.parameter_from_sqa,
    ),
    (
        "RiskMeasure",
        get_risk_measure,
        Encoder.risk_measure_to_sqa,
        Decoder.metric_from_sqa,
    ),
    (
        "RobustSearchSpace",
        get_robust_search_space,
        # Using generic search space methods for full-stack test coverage.
        Encoder.search_space_to_sqa,
        Decoder.search_space_from_sqa,
    ),
    (
        "RobustSearchSpace",
        get_robust_search_space_environmental,
        # Using generic search space methods for full-stack test coverage.
        Encoder.search_space_to_sqa,
        Decoder.search_space_from_sqa,
    ),
    (
        "ScalarizedObjective",
        get_scalarized_objective,
        Encoder.objective_to_sqa,
        Decoder.metric_from_sqa,
    ),
    (
        "ScalarizedOutcomeConstraint",
        get_scalarized_outcome_constraint,
        Encoder.outcome_constraint_to_sqa,
        Decoder.metric_from_sqa,
    ),
    (
        "SyntheticRunner",
        get_synthetic_runner,
        Encoder.runner_to_sqa,
        Decoder.runner_from_sqa,
    ),
    (
        "SumConstraint",
        get_sum_constraint1,
        Encoder.parameter_constraint_to_sqa,
        Decoder.parameter_constraint_from_sqa,
    ),
    (
        "SumConstraint",
        get_sum_constraint2,
        Encoder.parameter_constraint_to_sqa,
        Decoder.parameter_constraint_from_sqa,
    ),
    ("Trial", get_trial, Encoder.trial_to_sqa, Decoder.trial_from_sqa),
]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from logging import Logger
from time import time
from typing import List, Optional, TYPE_CHECKING

import pandas as pd

from ax.core.observation import Observation, ObservationFeatures
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import RobustSearchSpace, SearchSpace
from ax.exceptions.core import UnsupportedError
from ax.modelbridge.transforms.base import Transform
from ax.models.types import TConfig
from ax.utils.common.logger import get_logger
from ax.utils.common.timeutils import unixtime_to_pandas_ts
from ax.utils.common.typeutils import checked_cast, not_none

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import modelbridge as modelbridge_module  # noqa F401


logger: Logger = get_logger(__name__)


class TimeAsFeature(Transform):
    """Convert start time and duration into features that can be used for modeling.

    If no end_time is available, the current time is used.

    Duration is normalized to the unit cube.

    Transform is done in-place.

    TODO: revise this when better support for non-tunable features is added.
    """

    def __init__(
        self,
        search_space: Optional[SearchSpace] = None,
        observations: Optional[List[Observation]] = None,
        modelbridge: Optional["modelbridge_module.base.ModelBridge"] = None,
        config: Optional[TConfig] = None,
    ) -> None:
        assert observations is not None, "TimeAsFeature requires observations"
        if isinstance(search_space, RobustSearchSpace):
            raise UnsupportedError(
                "TimeAsFeature transform is not supported for RobustSearchSpace."
            )
        self.min_start_time: float = float("inf")
        self.max_start_time: float = float("-inf")
        self.min_duration: float = float("inf")
        self.max_duration: float = float("-inf")
        self.current_time: float = time()
        for obs in observations:
            obsf = obs.features
            if obsf.start_time is None:
                raise ValueError(
                    "Unable to use TimeAsFeature since not all observations have "
                    "start time specified."
                )
            start_time = not_none(obsf.start_time).timestamp()
            self.min_start_time = min(self.min_start_time, start_time)
            self.max_start_time = max(self.max_start_time, start_time)
            duration = self._get_duration(start_time=start_time, end_time=obsf.end_time)
            self.min_duration = min(self.min_duration, duration)
            self.max_duration = max(self.max_duration, duration)
            self.duration_range: float = self.max_duration - self.min_duration
            if self.duration_range == 0:
                # no need to case-distinguish during normalization
                self.duration_range = 1.0

    def _get_duration(
        self, start_time: float, end_time: Optional[pd.Timestamp]
    ) -> float:
        return (
            self.current_time if end_time is None else end_time.timestamp()
        ) - start_time

    def transform_observation_features(
        self, observation_features: List[ObservationFeatures]
    ) -> List[ObservationFeatures]:
        for obsf in observation_features:
            if obsf.start_time is not None:
                start_time = obsf.start_time.timestamp()
                obsf.parameters["start_time"] = start_time
                duration = self._get_duration(
                    start_time=start_time, end_time=obsf.end_time
                )
                # normalize duration to the unit cube
                obsf.parameters["duration"] = (
                    duration - self.min_duration
                ) / self.duration_range
            else:
                # start time can be None for pending arms that generated
                # with a model that did not use the TimeAsFeature transform.
                # In that case, assume the arm is going to be evaluated at the
                # current time, and that the duration is the midpoint of the
                # range.
                obsf.parameters["start_time"] = self.current_time
                obsf.parameters["duration"] = 0.5
        return observation_features

    def _transform_search_space(self, search_space: SearchSpace) -> SearchSpace:
        for p_name in ("start_time", "duration"):
            if p_name in search_space.parameters:
                raise ValueError(
                    f"Parameter name {p_name} is reserved when using "
                    "TimeAsFeature transform, but is part of the provided "
                    "search space. Please choose a different name for "
                    "this parameter."
                )
        param = RangeParameter(
            name="start_time",
            parameter_type=ParameterType.FLOAT,
            lower=self.min_start_time,
            upper=self.max_start_time,
        )
        search_space.add_parameter(param)
        param = RangeParameter(
            name="duration",
            parameter_type=ParameterType.FLOAT,
            # duration is normalized to [0,1]
            lower=0.0,
            upper=1.0,
        )
        search_space.add_parameter(param)
        return search_space

    def untransform_observation_features(
        self, observation_features: List[ObservationFeatures]
    ) -> List[ObservationFeatures]:
        for obsf in observation_features:
            start_time = checked_cast(float, obsf.parameters.pop("start_time"))
            obsf.start_time = unixtime_to_pandas_ts(start_time)
            obsf.end_time = unixtime_to_pandas_ts(
                checked_cast(float, obsf.parameters.pop("duration"))
                * self.duration_range
                + self.min_duration
                + start_time
            )
        return observation_features



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from dataclasses import dataclass
from math import inf
from typing import Dict, List, Optional, Tuple

from ax.core.base_trial import TrialStatus

from ax.core.experiment import Experiment
from ax.core.map_metric import MapMetric
from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    ParameterType,
    RangeParameter,
)
from ax.core.search_space import HierarchicalSearchSpace, SearchSpace
from ax.core.utils import get_model_times
from ax.telemetry.common import INITIALIZATION_MODELS, OTHER_MODELS

INITIALIZATION_MODEL_STRS: List[str] = [enum.value for enum in INITIALIZATION_MODELS]
OTHER_MODEL_STRS: List[str] = [enum.value for enum in OTHER_MODELS] + [None]


@dataclass(frozen=True)
class ExperimentCreatedRecord:
    """
    Record of the Experiment creation event. This can be used for telemetry in settings
    where many Experiments are being created either manually or programatically. In
    order to facilitate easy serialization only include simple types: numbers, strings,
    bools, and None.
    """

    experiment_name: Optional[str]
    experiment_type: Optional[str]

    # SearchSpace info
    num_continuous_range_parameters: int

    # Note: ordered ChoiceParameters and int RangeParameters should both utilize the
    # following fields
    num_int_range_parameters_small: int  # 2 - 3 elements
    num_int_range_parameters_medium: int  # 4 - 7 elements
    num_int_range_parameters_large: int  # 8 or more elements

    # Any RangeParameter can specify log space sampling
    num_log_scale_range_parameters: int

    num_unordered_choice_parameters_small: int  # 2 - 3 elements
    num_unordered_choice_parameters_medium: int  # 4 - 7 elements
    num_unordered_choice_parameters_large: int  # 8 or more elements

    num_fixed_parameters: int

    dimensionality: int
    hierarchical_tree_height: int  # Height of tree for HSS, 1 for base SearchSpace
    num_parameter_constraints: int

    # OptimizationConfig info
    num_objectives: int
    num_tracking_metrics: int
    num_outcome_constraints: int  # Includes ObjectiveThresholds in MOO

    # General Metrics info
    num_map_metrics: int
    metric_cls_to_quantity: Dict[str, int]

    # Runner info
    runner_cls: str

    @classmethod
    def from_experiment(cls, experiment: Experiment) -> ExperimentCreatedRecord:
        (
            num_continuous_range_parameters,
            num_int_range_parameters_small,
            num_int_range_parameters_medium,
            num_int_range_parameters_large,
            num_log_scale_range_parameters,
            num_unordered_choice_parameters_small,
            num_unordered_choice_parameters_medium,
            num_unordered_choice_parameters_large,
            num_fixed_parameters,
        ) = cls._get_param_counts_from_search_space(
            search_space=experiment.search_space
        )

        return cls(
            experiment_name=experiment._name,
            experiment_type=experiment._experiment_type,
            num_continuous_range_parameters=num_continuous_range_parameters,
            num_int_range_parameters_small=num_int_range_parameters_small,
            num_int_range_parameters_medium=num_int_range_parameters_medium,
            num_int_range_parameters_large=num_int_range_parameters_large,
            num_log_scale_range_parameters=num_log_scale_range_parameters,
            num_unordered_choice_parameters_small=num_unordered_choice_parameters_small,
            num_unordered_choice_parameters_medium=(
                num_unordered_choice_parameters_medium
            ),
            num_unordered_choice_parameters_large=num_unordered_choice_parameters_large,
            num_fixed_parameters=num_fixed_parameters,
            dimensionality=sum(
                1 for param in experiment.parameters.values() if param.cardinality() > 1
            ),
            hierarchical_tree_height=(
                experiment.search_space.height
                if isinstance(experiment.search_space, HierarchicalSearchSpace)
                else 1
            ),
            num_parameter_constraints=len(
                experiment.search_space.parameter_constraints
            ),
            num_objectives=(
                len(experiment.optimization_config.objective.metrics)
                if experiment.optimization_config is not None
                else 0
            ),
            num_tracking_metrics=len(experiment.tracking_metrics),
            num_outcome_constraints=(
                len(experiment.optimization_config.outcome_constraints)
                if experiment.optimization_config is not None
                else 0
            ),
            num_map_metrics=sum(
                1
                for metric in experiment.metrics.values()
                if isinstance(metric, MapMetric)
            ),
            metric_cls_to_quantity={
                cls_name: sum(
                    1
                    for metric in experiment.metrics.values()
                    if metric.__class__.__name__ == cls_name
                )
                for cls_name in {
                    metric.__class__.__name__ for metric in experiment.metrics.values()
                }
            },
            runner_cls=experiment.runner.__class__.__name__,
        )

    @staticmethod
    def _get_param_counts_from_search_space(
        search_space: SearchSpace,
    ) -> Tuple[int, int, int, int, int, int, int, int, int]:
        """
        Return counts of different types of parameters.

        returns:
            num_continuous_range_parameters

            num_int_range_parameters_small
            num_int_range_parameters_medium
            num_int_range_parameters_large

            num_log_scale_range_parameters

            num_unordered_choice_parameters_small
            num_unordered_choice_parameters_medium
            num_unordered_choice_parameters_large

            num_fixed_parameters
        """

        num_continuous_range_parameters = sum(
            1
            for param in search_space.parameters.values()
            if isinstance(param, RangeParameter)
            and param.parameter_type == ParameterType.FLOAT
        )
        num_int_range_parameters_small = sum(
            1
            for param in search_space.parameters.values()
            if (
                isinstance(param, RangeParameter)
                or (isinstance(param, ChoiceParameter) and param.is_ordered)
            )
            and (1.0 < param.cardinality() <= 3.0)
        )
        num_int_range_parameters_medium = sum(
            1
            for param in search_space.parameters.values()
            if (
                isinstance(param, RangeParameter)
                or (isinstance(param, ChoiceParameter) and param.is_ordered)
            )
            and (3.0 < param.cardinality() <= 7.0)
        )
        num_int_range_parameters_large = sum(
            1
            for param in search_space.parameters.values()
            if (
                isinstance(param, RangeParameter)
                or (isinstance(param, ChoiceParameter) and param.is_ordered)
            )
            and (7.0 < param.cardinality() < inf)
        )
        num_log_scale_range_parameters = sum(
            1
            for param in search_space.parameters.values()
            if isinstance(param, RangeParameter) and param.log_scale
        )
        num_unordered_choice_parameters_small = sum(
            1
            for param in search_space.parameters.values()
            if (isinstance(param, ChoiceParameter) and not param.is_ordered)
            and (1 < param.cardinality() <= 3)
        )
        num_unordered_choice_parameters_medium = sum(
            1
            for param in search_space.parameters.values()
            if (isinstance(param, ChoiceParameter) and not param.is_ordered)
            and (3 < param.cardinality() <= 7)
        )
        num_unordered_choice_parameters_large = sum(
            1
            for param in search_space.parameters.values()
            if (isinstance(param, ChoiceParameter) and not param.is_ordered)
            and param.cardinality() > 7
        )
        num_fixed_parameters = sum(
            1
            for param in search_space.parameters.values()
            if isinstance(param, FixedParameter)
        )

        return (
            num_continuous_range_parameters,
            num_int_range_parameters_small,
            num_int_range_parameters_medium,
            num_int_range_parameters_large,
            num_log_scale_range_parameters,
            num_unordered_choice_parameters_small,
            num_unordered_choice_parameters_medium,
            num_unordered_choice_parameters_large,
            num_fixed_parameters,
        )


@dataclass(frozen=True)
class ExperimentCompletedRecord:
    """
    Record of the Experiment completion event. This can be used for telemetry in
    settings where many Experiments are being created either manually or
    programatically. In order to facilitate easy serialization only include simple
    types: numbers, strings, bools, and None.
    """

    num_initialization_trials: int
    num_bayesopt_trials: int
    num_other_trials: int

    num_completed_trials: int
    num_failed_trials: int
    num_abandoned_trials: int
    num_early_stopped_trials: int

    total_fit_time: int
    total_gen_time: int

    @classmethod
    def from_experiment(cls, experiment: Experiment) -> ExperimentCompletedRecord:
        trial_count_by_status = {
            status: len(trials)
            for status, trials in experiment.trials_by_status.items()
        }

        model_keys = [
            [gr._model_key for gr in trial.generator_runs]
            for trial in experiment.trials.values()
        ]

        fit_time, gen_time = get_model_times(experiment=experiment)

        return cls(
            num_initialization_trials=sum(
                1
                for model_key_list in model_keys
                if all(
                    model_key in INITIALIZATION_MODEL_STRS
                    for model_key in model_key_list
                )
            ),
            num_bayesopt_trials=sum(
                1
                for model_key_list in model_keys
                if any(
                    model_key not in INITIALIZATION_MODEL_STRS
                    and model_key not in OTHER_MODEL_STRS
                    for model_key in model_key_list
                )
            ),
            num_other_trials=sum(
                1
                for model_key_list in model_keys
                if all(model_key in OTHER_MODEL_STRS for model_key in model_key_list)
            ),
            num_completed_trials=trial_count_by_status[TrialStatus.COMPLETED],
            num_failed_trials=trial_count_by_status[TrialStatus.FAILED],
            num_abandoned_trials=trial_count_by_status[TrialStatus.ABANDONED],
            num_early_stopped_trials=trial_count_by_status[TrialStatus.EARLY_STOPPED],
            total_fit_time=int(fit_time),
            total_gen_time=int(gen_time),
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from typing import Any, Callable, List, Optional

import numpy as np
import pandas as pd
from ax.core.base_trial import BaseTrial
from ax.core.data import Data
from ax.core.metric import Metric, MetricFetchE, MetricFetchResult
from ax.core.types import TParameterization
from ax.utils.common.result import Err, Ok


class NoisyFunctionMetric(Metric):
    """A metric defined by a generic deterministic function, with normal noise
    with mean 0 and mean_sd scale added to the result.
    """

    def __init__(
        self,
        name: str,
        param_names: List[str],
        noise_sd: Optional[float] = 0.0,
        lower_is_better: Optional[bool] = None,
    ) -> None:
        """
        Metric is computed by evaluating a deterministic function, implemented
        in the `f` method defined on this class.

        f will expect an array x, which is constructed from the arm
        parameters by extracting the values of the parameter names given in
        param_names, in that order.

        Args:
            name: Name of the metric
            param_names: An ordered list of names of parameters to be passed
                to the deterministic function.
            noise_sd: Scale of normal noise added to the function result. If
                None, interpret the function as noisy with unknown noise level.
            lower_is_better: Flag for metrics which should be minimized.
        """
        self.param_names = param_names
        self.noise_sd = noise_sd
        super().__init__(name=name, lower_is_better=lower_is_better)

    @classmethod
    def is_available_while_running(cls) -> bool:
        return True

    def clone(self) -> NoisyFunctionMetric:
        return self.__class__(
            name=self._name,
            param_names=self.param_names,
            noise_sd=self.noise_sd,
            lower_is_better=self.lower_is_better,
        )

    def fetch_trial_data(
        self, trial: BaseTrial, noisy: bool = True, **kwargs: Any
    ) -> MetricFetchResult:
        try:
            noise_sd = self.noise_sd if noisy else 0.0
            arm_names = []
            mean = []
            for name, arm in trial.arms_by_name.items():
                arm_names.append(name)
                val = self._evaluate(params=arm.parameters)
                if noise_sd:
                    val = val + noise_sd * np.random.randn()
                mean.append(val)
            # indicate unknown noise level in data
            if noise_sd is None:
                noise_sd = float("nan")
            df = pd.DataFrame(
                {
                    "arm_name": arm_names,
                    "metric_name": self.name,
                    "mean": mean,
                    "sem": noise_sd,
                    "trial_index": trial.index,
                    "n": 10000 / len(arm_names),
                    "frac_nonnull": mean,
                }
            )

            return Ok(value=Data(df=df))

        except Exception as e:
            return Err(
                MetricFetchE(message=f"Failed to fetch {self.name}", exception=e)
            )

    def _evaluate(self, params: TParameterization) -> float:
        x = np.array([params[p] for p in self.param_names])
        return self.f(x)

    def f(self, x: np.ndarray) -> float:
        """The deterministic function that produces the metric outcomes."""
        raise NotImplementedError


class GenericNoisyFunctionMetric(NoisyFunctionMetric):
    def __init__(
        self,
        name: str,
        f: Callable[[TParameterization], float],
        noise_sd: Optional[float] = 0.0,
        lower_is_better: Optional[bool] = None,
    ) -> None:
        """
        Metric is computed by evaluating a deterministic function, implemented in f.

        Args:
            name: Name of the metric.
            f: A callable accepting a dictionary from parameter names to
                values and returning a float metric value.
            noise_sd: Scale of normal noise added to the function result. If
                None, interpret the function as noisy with unknown noise level.
            lower_is_better: Flag for metrics which should be minimized.

        Note: Since this metric setup uses a generic callable it cannot be serialized
        and will not play well with storage.
        """
        self._f = f
        self.noise_sd = noise_sd
        Metric.__init__(self, name=name, lower_is_better=lower_is_better)

    @property
    def param_names(self) -> List[str]:
        raise NotImplementedError(
            "GenericNoisyFunctionMetric does not implement a param_names attribute"
        )

    def _evaluate(self, params: TParameterization) -> float:
        return self._f(params)

    def clone(self) -> GenericNoisyFunctionMetric:
        return self.__class__(
            name=self._name,
            f=self._f,
            noise_sd=self.noise_sd,
            lower_is_better=self.lower_is_better,
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from abc import ABCMeta, abstractmethod, abstractproperty
from copy import deepcopy
from enum import Enum
from math import inf
from typing import cast, Dict, List, Optional, Tuple, Type, Union
from warnings import warn

from ax.core.types import TNumeric, TParamValue, TParamValueList
from ax.exceptions.core import AxParameterWarning, UserInputError
from ax.utils.common.base import SortableBase
from ax.utils.common.typeutils import not_none
from pyre_extensions import assert_is_instance

# Tolerance for floating point comparisons. This is relatively permissive,
# and allows for serializing at rather low numerical precision.
# TODO: Do a more comprehensive audit of how floating point precision issues
# may creep up and implement a more principled fix
EPS = 1.5e-7
MAX_VALUES_CHOICE_PARAM = 1000
FIXED_CHOICE_PARAM_ERROR = (
    "ChoiceParameters require multiple feasible values. "
    "Please use FixedParameter instead when setting a single possible value."
)
REPR_FLAGS_IF_TRUE_ONLY = [
    "is_fidelity",
    "is_task",
    "is_hierarchical",
    "log_scale",
    "logit_scale",
]


class ParameterType(Enum):
    BOOL: int = 0
    INT: int = 1
    FLOAT: int = 2
    STRING: int = 3

    @property
    def is_numeric(self) -> bool:
        return self == ParameterType.INT or self == ParameterType.FLOAT


TParameterType = Union[Type[int], Type[float], Type[str], Type[bool]]

# pyre: PARAMETER_PYTHON_TYPE_MAP is declared to have type
# pyre: `Dict[ParameterType, Union[Type[bool], Type[float], Type[int],
# pyre: Type[str]]]` but is used as type `Dict[ParameterType,
# pyre-fixme[9]: Type[Union[float, str]]]`.
PARAMETER_PYTHON_TYPE_MAP: Dict[ParameterType, TParameterType] = {
    ParameterType.INT: int,
    ParameterType.FLOAT: float,
    ParameterType.STRING: str,
    ParameterType.BOOL: bool,
}

SUPPORTED_PARAMETER_TYPES: Tuple[
    Union[Type[bool], Type[float], Type[int], Type[str]], ...
] = tuple(PARAMETER_PYTHON_TYPE_MAP.values())


# pyre-fixme[24]: Generic type `type` expects 1 type parameter, use `typing.Type` to
#  avoid runtime subscripting errors.
def _get_parameter_type(python_type: Type) -> ParameterType:
    """Given a Python type, retrieve corresponding Ax ``ParameterType``."""
    for param_type, py_type in PARAMETER_PYTHON_TYPE_MAP.items():
        if py_type == python_type:
            return param_type
    raise ValueError(f"No Ax parameter type corresponding to {python_type}.")


class Parameter(SortableBase, metaclass=ABCMeta):
    _is_fidelity: bool = False
    _name: str
    _target_value: TParamValue = None
    _parameter_type: ParameterType

    def cast(self, value: TParamValue) -> TParamValue:
        if value is None:
            return None
        return self.python_type(value)

    @abstractmethod
    def validate(self, value: TParamValue) -> bool:
        pass

    @abstractmethod
    def cardinality(self) -> float:
        pass

    @property
    def python_type(self) -> TParameterType:
        """The python type for the corresponding ParameterType enum.

        Used primarily for casting values of unknown type to conform
        to that of the parameter.
        """
        return PARAMETER_PYTHON_TYPE_MAP[self.parameter_type]

    def is_valid_type(self, value: TParamValue) -> bool:
        """Whether a given value's type is allowed by this parameter."""
        return type(value) is self.python_type

    @property
    def is_numeric(self) -> bool:
        return self.parameter_type.is_numeric

    @property
    def is_fidelity(self) -> bool:
        return self._is_fidelity

    @property
    def is_hierarchical(self) -> bool:
        return isinstance(self, (ChoiceParameter, FixedParameter)) and bool(
            self._dependents
        )

    @property
    def target_value(self) -> TParamValue:
        return self._target_value

    @property
    def parameter_type(self) -> ParameterType:
        return self._parameter_type

    @property
    def name(self) -> str:
        return self._name

    @property
    def dependents(self) -> Dict[TParamValue, List[str]]:
        raise NotImplementedError(
            "Only choice hierarchical parameters are currently supported."
        )

    def clone(self) -> Parameter:
        # pyre-fixme[7]: Expected `Parameter` but got implicit return value of `None`.
        pass

    @property
    def _unique_id(self) -> str:
        return str(self)

    def _base_repr(self) -> str:
        ret_val = (
            f"{self.__class__.__name__}("
            f"name='{self._name}', "
            f"parameter_type={self.parameter_type.name}, "
            f"{self.domain_repr}"
        )

        # Add binary flags.
        for flag in self.available_flags:
            val = getattr(self, flag, False)
            if flag not in REPR_FLAGS_IF_TRUE_ONLY or (
                flag in REPR_FLAGS_IF_TRUE_ONLY and val is True
            ):
                ret_val += f", {flag}={val}"

        # Add target_value if one exists.
        if self.target_value is not None:
            tval_rep = self.target_value
            if self.parameter_type == ParameterType.STRING:
                tval_rep = f"'{tval_rep}'"
            ret_val += f", target_value={tval_rep}"

        return ret_val

    @abstractproperty
    def domain_repr(self) -> str:
        """Returns a string representation of the domain."""
        pass

    @property
    def available_flags(self) -> List[str]:
        """List of boolean attributes that can be set on this parameter."""
        return ["is_fidelity"]

    @property
    def summary_dict(
        self,
    ) -> Dict[str, Union[TParamValueList, TParamValue, str, List[str]]]:

        # Assemble dict.
        summary_dict = {
            "name": self.name,
            "type": self.__class__.__name__.removesuffix("Parameter"),
            "domain": self.domain_repr,
            "parameter_type": self.parameter_type.name.lower(),
        }

        # Extract flags.
        flags = []
        for flag in self.available_flags:
            flag_val = getattr(self, flag, None)
            flag_repr = flag.removeprefix("is_")
            if flag == "sort_values":
                flag_repr = "sorted"
            if flag_val is True:
                flags.append(flag_repr)
            elif flag_val is False and flag not in REPR_FLAGS_IF_TRUE_ONLY:
                flags.append("un" + flag_repr)

        # Add flags, target_values, and dependents if present.
        if flags:
            summary_dict["flags"] = ", ".join(flags)
        if getattr(self, "is_fidelity", False) or getattr(self, "is_task", False):
            summary_dict["target_value"] = self.target_value
        if getattr(self, "is_hierarchical", False):
            summary_dict["dependents"] = self.dependents

        return summary_dict


class RangeParameter(Parameter):
    """Parameter object that specifies a range of values."""

    def __init__(
        self,
        name: str,
        parameter_type: ParameterType,
        lower: float,
        upper: float,
        log_scale: bool = False,
        logit_scale: bool = False,
        digits: Optional[int] = None,
        is_fidelity: bool = False,
        target_value: TParamValue = None,
    ) -> None:
        """Initialize RangeParameter

        Args:
            name: Name of the parameter.
            parameter_type: Enum indicating the type of parameter
                value (e.g. string, int).
            lower: Lower bound of the parameter range (inclusive).
            upper: Upper bound of the parameter range (inclusive).
            log_scale: Whether to sample in the log space when drawing
                random values of the parameter.
            logit_scale: Whether to sample in logit space when drawing
                random values of the parameter.
            digits: Number of digits to round values to for float type.
            is_fidelity: Whether this parameter is a fidelity parameter.
            target_value: Target value of this parameter if it is a fidelity.
        """
        if is_fidelity and (target_value is None):
            raise UserInputError(
                "`target_value` should not be None for the fidelity parameter: "
                "{}".format(name)
            )

        self._name = name
        if parameter_type not in (ParameterType.INT, ParameterType.FLOAT):
            raise UserInputError("RangeParameter type must be int or float.")
        self._parameter_type = parameter_type
        self._digits = digits
        self._lower: TNumeric = not_none(self.cast(lower))
        self._upper: TNumeric = not_none(self.cast(upper))
        self._log_scale = log_scale
        self._logit_scale = logit_scale
        self._is_fidelity = is_fidelity
        self._target_value: Optional[TNumeric] = self.cast(target_value)

        self._validate_range_param(
            parameter_type=parameter_type,
            lower=lower,
            upper=upper,
            log_scale=log_scale,
            logit_scale=logit_scale,
        )

    def cardinality(self) -> TNumeric:
        if self.parameter_type == ParameterType.FLOAT:
            return inf

        return int(self.upper) - int(self.lower) + 1

    def _validate_range_param(
        self,
        lower: TNumeric,
        upper: TNumeric,
        log_scale: bool,
        logit_scale: bool,
        parameter_type: Optional[ParameterType] = None,
    ) -> None:
        if parameter_type and parameter_type not in (
            ParameterType.INT,
            ParameterType.FLOAT,
        ):
            raise UserInputError("RangeParameter type must be int or float.")

        upper = float(upper)
        if lower >= upper:
            raise UserInputError(
                f"Upper bound of {self.name} must be strictly larger than lower."
                f"Got: ({lower}, {upper})."
            )
        width: float = upper - lower
        if width < 100 * EPS:
            raise UserInputError(
                f"Parameter range ({width}) is very small and likely "
                "to cause numerical errors. Consider reparameterizing your "
                "problem by scaling the parameter."
            )
        if log_scale and logit_scale:
            raise UserInputError("Can't use both log and logit.")
        if log_scale and lower <= 0:
            raise UserInputError("Cannot take log when min <= 0.")
        if logit_scale and (lower <= 0 or upper >= 1):
            raise UserInputError("Logit requires lower > 0 and upper < 1")
        if not (self.is_valid_type(lower)) or not (self.is_valid_type(upper)):
            raise UserInputError(
                f"[{lower}, {upper}] is an invalid range for this parameter."
            )

    @property
    def upper(self) -> TNumeric:
        """Upper bound of the parameter range.

        Value is cast to parameter type upon set and also validated
        to ensure the bound is strictly greater than lower bound.
        """
        return self._upper

    @upper.setter
    def upper(self, value: TNumeric) -> None:
        self._validate_range_param(
            lower=self.lower,
            upper=value,
            log_scale=self.log_scale,
            logit_scale=self.logit_scale,
        )
        self._upper = not_none(self.cast(value))

    @property
    def lower(self) -> TNumeric:
        """Lower bound of the parameter range.

        Value is cast to parameter type upon set and also validated
        to ensure the bound is strictly less than upper bound.
        """
        return self._lower

    @lower.setter
    def lower(self, value: TNumeric) -> None:
        self._validate_range_param(
            lower=value,
            upper=self.upper,
            log_scale=self.log_scale,
            logit_scale=self.logit_scale,
        )
        self._lower = not_none(self.cast(value))

    @property
    def digits(self) -> Optional[int]:
        """Number of digits to round values to for float type.

        Upper and lower bound are re-cast after this property is changed.
        """
        return self._digits

    @property
    def log_scale(self) -> bool:
        """Whether the parameter's random values should be sampled from log space."""
        return self._log_scale

    @property
    def logit_scale(self) -> bool:
        """Whether the parameter's random values should be sampled from logit space."""
        return self._logit_scale

    def update_range(
        self, lower: Optional[float] = None, upper: Optional[float] = None
    ) -> RangeParameter:
        """Set the range to the given values.

        If lower or upper is not provided, it will be left at its current value.

        Args:
            lower: New value for the lower bound.
            upper: New value for the upper bound.
        """
        if lower is None:
            lower = self._lower
        if upper is None:
            upper = self._upper

        cast_lower = not_none(self.cast(lower))
        cast_upper = not_none(self.cast(upper))
        self._validate_range_param(
            lower=cast_lower,
            upper=cast_upper,
            log_scale=self.log_scale,
            logit_scale=self.logit_scale,
        )
        self._lower = cast_lower
        self._upper = cast_upper
        return self

    def set_digits(self, digits: int) -> RangeParameter:
        self._digits = digits

        # Re-scale min and max to new digits definition
        cast_lower = not_none(self.cast(self._lower))
        cast_upper = not_none(self.cast(self._upper))
        if cast_lower >= cast_upper:
            raise UserInputError(
                f"Lower bound {cast_lower} is >= upper bound {cast_upper}."
            )

        self._lower = cast_lower
        self._upper = cast_upper
        return self

    def set_log_scale(self, log_scale: bool) -> RangeParameter:
        self._log_scale = log_scale
        return self

    def set_logit_scale(self, logit_scale: bool) -> RangeParameter:
        self._logit_scale = logit_scale
        return self

    def validate(self, value: TParamValue, tol: float = EPS) -> bool:
        """Returns True if input is a valid value for the parameter.

        Checks that value is of the right type and within
        the valid range for the parameter. Returns False if value is None.

        Args:
            value: Value being checked.
            tol: Absolute tolerance for floating point comparisons.

        Returns:
            True if valid, False otherwise.
        """
        if value is None:
            return False

        if not self.is_valid_type(value):
            return False
        # pyre-fixme[58]: `>=` is not supported for operand types `Union[bool,
        #  float, int, str]` and `float`.
        # pyre-fixme[58]: `<=` is not supported for operand types `Union[bool,
        #  float, int, str]` and `float`.
        return value >= self._lower - tol and value <= self._upper + tol

    def is_valid_type(self, value: TParamValue) -> bool:
        """Same as default except allows floats whose value is an int
        for Int parameters.
        """
        if not (isinstance(value, float) or isinstance(value, int)):
            return False

        # This might have issues with ints > 2^24
        if self.parameter_type is ParameterType.INT:
            return isinstance(value, int) or float(not_none(value)).is_integer()
        return True

    def clone(self) -> RangeParameter:
        return RangeParameter(
            name=self._name,
            parameter_type=self._parameter_type,
            lower=self._lower,
            upper=self._upper,
            log_scale=self._log_scale,
            logit_scale=self._logit_scale,
            digits=self._digits,
            is_fidelity=self._is_fidelity,
            target_value=self._target_value,
        )

    def cast(self, value: TParamValue) -> Optional[TNumeric]:
        if value is None:
            return None
        if self.parameter_type is ParameterType.FLOAT and self._digits is not None:
            return round(float(value), not_none(self._digits))
        return assert_is_instance(self.python_type(value), TNumeric)

    def __repr__(self) -> str:
        ret_val = self._base_repr()

        if self._digits is not None:
            ret_val += f", digits={self._digits}"

        return ret_val + ")"

    @property
    def available_flags(self) -> List[str]:
        """List of boolean attributes that can be set on this parameter."""
        return super().available_flags + ["log_scale", "logit_scale"]

    @property
    def domain_repr(self) -> str:
        """Returns a string representation of the domain."""
        return f"range={[self.lower, self.upper]}"


class ChoiceParameter(Parameter):
    """Parameter object that specifies a discrete set of values.

    Args:
        name: Name of the parameter.
        parameter_type: Enum indicating the type of parameter
            value (e.g. string, int).
        values: List of allowed values for the parameter.
        is_ordered: If False, the parameter is a categorical variable.
            Defaults to False if parameter_type is STRING and ``values``
            is longer than 2, else True.
        is_task: Treat the parameter as a task parameter for modeling.
        is_fidelity: Whether this parameter is a fidelity parameter.
        target_value: Target value of this parameter if it's a fidelity or
            task parameter.
        sort_values: Whether to sort ``values`` before encoding.
            Defaults to False if ``parameter_type`` is STRING, else
            True.
        dependents: Optional mapping for parameters in hierarchical search
            spaces; format is { value -> list of dependent parameter names }.
    """

    def __init__(
        self,
        name: str,
        parameter_type: ParameterType,
        values: List[TParamValue],
        is_ordered: Optional[bool] = None,
        is_task: bool = False,
        is_fidelity: bool = False,
        target_value: TParamValue = None,
        sort_values: Optional[bool] = None,
        dependents: Optional[Dict[TParamValue, List[str]]] = None,
    ) -> None:
        if (is_fidelity or is_task) and (target_value is None):
            ptype = "fidelity" if is_fidelity else "task"
            raise UserInputError(
                f"`target_value` should not be None for the {ptype} parameter: "
                "{}".format(name)
            )

        self._name = name
        self._parameter_type = parameter_type
        self._is_task = is_task
        self._is_fidelity = is_fidelity
        self._target_value: TParamValue = self.cast(target_value)
        # A choice parameter with only one value is a FixedParameter.
        if not len(values) > 1:
            raise UserInputError(f"{self._name}({values}): {FIXED_CHOICE_PARAM_ERROR}")
        # Cap the number of possible values.
        if len(values) > MAX_VALUES_CHOICE_PARAM:
            raise UserInputError(
                f"`ChoiceParameter` with more than {MAX_VALUES_CHOICE_PARAM} values "
                "is not supported! Use a `RangeParameter` instead."
            )
        # Remove duplicate values.
        # Using dict to deduplicate here since set doesn't preserve order but dict does.
        if len(values) != len(dict_values := dict.fromkeys(values)):
            warn(
                f"Duplicate values found for ChoiceParameter {name}. "
                "Initializing the parameter with duplicate values removed. ",
                AxParameterWarning,
                stacklevel=2,
            )
            values = list(dict_values)

        if is_ordered is False and len(values) == 2:
            is_ordered = True
            warn(
                f"Changing `is_ordered` to `True` for `ChoiceParameter` '{name}' since "
                "there are only two possible values.",
                AxParameterWarning,
                stacklevel=3,
            )
        self._is_ordered: bool = (
            is_ordered
            if is_ordered is not None
            else self._get_default_is_ordered_and_warn(num_choices=len(values))
        )
        # sort_values defaults to True if the parameter is not a string
        self._sort_values: bool = (
            sort_values
            if sort_values is not None
            else self._get_default_sort_values_and_warn()
        )
        if self.sort_values:
            values = cast(List[TParamValue], sorted([not_none(v) for v in values]))
        self._values: List[TParamValue] = self._cast_values(values)

        if dependents:
            for value in dependents:
                if value not in self.values:
                    raise UserInputError(
                        f"Value {value} in `dependents` "
                        f"argument is not among the parameter values: {self.values}."
                    )
        # NOTE: We don't need to check that dependent parameters actually exist as
        # that is done in `HierarchicalSearchSpace` constructor.
        self._dependents = dependents

    def _get_default_is_ordered_and_warn(self, num_choices: int) -> bool:
        default_bool = self._parameter_type != ParameterType.STRING or num_choices == 2
        if self._parameter_type == ParameterType.STRING and num_choices > 2:
            motivation = " since the parameter is a string with more than 2 choices."
        elif num_choices == 2:
            motivation = " since there are exactly two choices."
        else:
            motivation = " since the parameter is not of type string."
        warn(
            f'`is_ordered` is not specified for `ChoiceParameter` "{self._name}". '
            f"Defaulting to `{default_bool}` {motivation}. To override this behavior "
            f"(or avoid this warning), specify `is_ordered` during `ChoiceParameter` "
            "construction. Note that choice parameters with exactly 2 choices are "
            "always considered ordered and that the user-supplied `is_ordered` has no "
            "effect in this particular case.",
            AxParameterWarning,
            stacklevel=3,
        )
        return default_bool

    def _get_default_sort_values_and_warn(self) -> bool:
        default_bool = self._parameter_type != ParameterType.STRING
        warn(
            f'`sort_values` is not specified for `ChoiceParameter` "{self._name}". '
            f"Defaulting to `{default_bool}` for parameters of `ParameterType` "
            f"{self.parameter_type.name}. To override this behavior (or avoid this "
            f"warning), specify `sort_values` during `ChoiceParameter` construction.",
            AxParameterWarning,
            stacklevel=3,
        )
        return default_bool

    def cardinality(self) -> float:
        return len(self.values)

    @property
    def sort_values(self) -> bool:
        return self._sort_values

    @property
    def is_ordered(self) -> bool:
        return self._is_ordered

    @property
    def is_task(self) -> bool:
        return self._is_task

    @property
    def values(self) -> List[TParamValue]:
        return self._values

    def set_values(self, values: List[TParamValue]) -> ChoiceParameter:
        """Set the list of allowed values for parameter.

        Cast all input values to the parameter type.

        Args:
            values: New list of allowed values.
        """
        # A choice parameter with only one value is a FixedParameter.
        if not len(values) > 1:
            raise UserInputError(FIXED_CHOICE_PARAM_ERROR)
        self._values = self._cast_values(values)
        return self

    def add_values(self, values: List[TParamValue]) -> ChoiceParameter:
        """Add input list to the set of allowed values for parameter.

        Cast all input values to the parameter type.

        Args:
            values: Values being added to the allowed list.
        """
        self._values.extend(self._cast_values(values))
        return self

    def validate(self, value: TParamValue) -> bool:
        """Checks that the input is in the list of allowed values.

        Args:
            value: Value being checked.

        Returns:
            True if valid, False otherwise.
        """
        return value in self._values

    @property
    def dependents(self) -> Dict[TParamValue, List[str]]:
        if not self.is_hierarchical:
            raise NotImplementedError(
                "Only hierarchical parameters support the `dependents` property."
            )
        return not_none(self._dependents)

    def _cast_values(self, values: List[TParamValue]) -> List[TParamValue]:
        return [self.cast(value) for value in values]

    def clone(self) -> ChoiceParameter:
        return ChoiceParameter(
            name=self._name,
            parameter_type=self._parameter_type,
            values=self._values,
            is_ordered=self._is_ordered,
            is_task=self._is_task,
            is_fidelity=self._is_fidelity,
            target_value=self._target_value,
            sort_values=self._sort_values,
            dependents=deepcopy(self._dependents),
        )

    def __repr__(self) -> str:
        ret_val = self._base_repr()

        if self._dependents:
            ret_val += f", dependents={self._dependents}"

        return ret_val + ")"

    @property
    def available_flags(self) -> List[str]:
        """List of boolean attributes that can be set on this parameter."""
        return super().available_flags + [
            "is_ordered",
            "is_hierarchical",
            "is_task",
            "sort_values",
        ]

    @property
    def domain_repr(self) -> str:
        """Returns a string representation of the domain."""
        return f"values={self.values}"


class FixedParameter(Parameter):
    """Parameter object that specifies a single fixed value."""

    def __init__(
        self,
        name: str,
        parameter_type: ParameterType,
        value: TParamValue,
        is_fidelity: bool = False,
        target_value: TParamValue = None,
        dependents: Optional[Dict[TParamValue, List[str]]] = None,
    ) -> None:
        """Initialize FixedParameter

        Args:
            name: Name of the parameter.
            parameter_type: Enum indicating the type of parameter
                value (e.g. string, int).
            value: The fixed value of the parameter.
            is_fidelity: Whether this parameter is a fidelity parameter.
            target_value: Target value of this parameter if it is a fidelity.
            dependents: Optional mapping for parameters in hierarchical search
                spaces; format is { value -> list of dependent parameter names }.
        """
        if is_fidelity and (target_value is None):
            raise UserInputError(
                "`target_value` should not be None for the fidelity parameter: "
                "{}".format(name)
            )

        self._name = name
        self._parameter_type = parameter_type
        self._value: TParamValue = self.cast(value)
        self._is_fidelity = is_fidelity
        self._target_value: TParamValue = self.cast(target_value)
        # NOTE: We don't need to check that dependent parameters actually exist as
        # that is done in `HierarchicalSearchSpace` constructor.
        if dependents:
            if len(dependents) > 1 or next(iter(dependents.keys())) != self.value:
                raise UserInputError(
                    "The only expected key in `dependents` for fixed parameter "
                    f"{self.name}: {self.value}; got: {dependents}."
                )
        self._dependents = dependents

    def cardinality(self) -> float:
        return 1.0

    @property
    def value(self) -> TParamValue:
        return self._value

    def set_value(self, value: TParamValue) -> FixedParameter:
        self._value = self.cast(value)
        return self

    def validate(self, value: TParamValue) -> bool:
        """Checks that the input is equal to the fixed value.

        Args:
            value: Value being checked.

        Returns:
            True if valid, False otherwise.
        """
        return value == self._value

    @property
    def dependents(self) -> Dict[TParamValue, List[str]]:
        if not self.is_hierarchical:
            raise NotImplementedError(
                "Only hierarchical parameters support the `dependents` property."
            )
        return not_none(self._dependents)

    def clone(self) -> FixedParameter:
        return FixedParameter(
            name=self._name,
            parameter_type=self._parameter_type,
            value=self._value,
            is_fidelity=self._is_fidelity,
            target_value=self._target_value,
            dependents=self._dependents,
        )

    def __repr__(self) -> str:
        ret_val = self._base_repr()
        return ret_val + ")"

    @property
    def available_flags(self) -> List[str]:
        """List of boolean attributes that can be set on this parameter."""
        return super().available_flags + ["is_hierarchical"]

    @property
    def domain_repr(self) -> str:
        """Returns a string representation of the domain."""
        if self._parameter_type == ParameterType.STRING:
            return f"value='{self._value}'"
        else:
            return f"value={self._value}"



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

# flake8: noqa F401
from ax.modelbridge import transforms
from ax.modelbridge.base import ModelBridge
from ax.modelbridge.factory import (
    get_factorial,
    get_sobol,
    get_thompson,
    get_uniform,
    Models,
)
from ax.modelbridge.map_torch import MapTorchModelBridge
from ax.modelbridge.torch import TorchModelBridge

__all__ = [
    "MapTorchModelBridge",
    "ModelBridge",
    "Models",
    "TorchModelBridge",
    "get_factorial",
    "get_sobol",
    "get_thompson",
    "get_uniform",
    "transforms",
]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from collections import defaultdict
from math import isnan
from numbers import Number
from typing import Any, Callable, Dict, List, Optional, Tuple, TYPE_CHECKING, Union

import numpy as np
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.optimization_config import OptimizationConfig
from ax.core.parameter import Parameter
from ax.core.parameter_constraint import ParameterConstraint
from ax.core.search_space import RobustSearchSpace, SearchSpace
from ax.modelbridge.transforms.derelativize import Derelativize
from scipy.stats import norm


if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import modelbridge as modelbridge_module  # noqa F401


# pyre-fixme[24]: Generic type `dict` expects 2 type parameters, use `typing.Dict`
#  to avoid runtime subscripting errors.
class ClosestLookupDict(dict):
    """A dictionary with numeric keys that looks up the closest key."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        # pyre-fixme[4]: Attribute must be annotated.
        self._keys = sorted(self.keys())

    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    def __setitem__(self, key: Number, val: Any) -> None:
        if not isinstance(key, Number):
            raise ValueError("ClosestLookupDict only allows numerical keys.")
        super().__setitem__(key, val)
        ipos = np.searchsorted(self._keys, key)
        self._keys.insert(ipos, key)

    # pyre-fixme[3]: Return annotation cannot be `Any`.
    def __getitem__(self, key: Number) -> Any:
        try:
            return super().__getitem__(key)
        except KeyError:
            if not self.keys():
                raise RuntimeError("ClosestLookupDict is empty.")
            ipos = np.searchsorted(self._keys, key)
            if ipos == 0:
                return super().__getitem__(self._keys[0])
            elif ipos == len(self._keys):
                return super().__getitem__(self._keys[-1])
            lkey, rkey = self._keys[ipos - 1 : ipos + 1]
            if np.abs(key - lkey) <= np.abs(key - rkey):  # pyre-ignore [58]
                return super().__getitem__(lkey)
            else:
                return super().__getitem__(rkey)


def get_data(
    observation_data: List[ObservationData], metric_names: Union[List[str], None] = None
) -> Dict[str, List[float]]:
    """Extract all metrics if `metric_names` is None."""
    Ys = defaultdict(list)
    for obsd in observation_data:
        for i, m in enumerate(obsd.metric_names):
            if metric_names is None or m in metric_names:
                Ys[m].append(obsd.means[i])
    return Ys


def match_ci_width_truncated(
    mean: float,
    variance: float,
    transform: Callable[[float], float],
    level: float = 0.95,
    margin: float = 0.001,
    lower_bound: float = 0.0,
    upper_bound: float = 1.0,
    clip_mean: bool = False,
) -> Tuple[float, float]:
    """Estimate a transformed variance using the match ci width method.

    See log_y transform for the original. Here, bounds are forced to lie
    within a [lower_bound, upper_bound] interval after transformation."""
    fac = norm.ppf(1 - (1 - level) / 2)
    d = fac * np.sqrt(variance)
    if clip_mean:
        mean = np.clip(mean, lower_bound + margin, upper_bound - margin)
    right = min(mean + d, upper_bound - margin)
    left = max(mean - d, lower_bound + margin)
    width_asym = transform(right) - transform(left)
    new_mean = transform(mean)
    new_variance = float("nan") if isnan(variance) else (width_asym / 2 / fac) ** 2
    return new_mean, new_variance


def construct_new_search_space(
    search_space: SearchSpace,
    parameters: List[Parameter],
    parameter_constraints: Optional[List[ParameterConstraint]] = None,
) -> SearchSpace:
    """Construct a search space with the transformed arguments.

    If the `search_space` is a `RobustSearchSpace`, this will use its
    environmental variables and distributions, and remove the environmental
    variables from `parameters` before constructing.

    Args:
        parameters: List of transformed parameter objects.
        parameter_constraints: List of parameter constraints.

    Returns:
        The new search space instance.
    """
    new_kwargs: Dict[str, Any] = {
        "parameters": parameters,
        "parameter_constraints": parameter_constraints,
    }
    if isinstance(search_space, RobustSearchSpace):
        env_vars = list(search_space._environmental_variables.values())
        if env_vars:
            # Add environmental variables and remove them from parameters.
            new_kwargs["environmental_variables"] = env_vars
            new_kwargs["parameters"] = [p for p in parameters if p not in env_vars]
        new_kwargs["parameter_distributions"] = search_space.parameter_distributions
        new_kwargs["num_samples"] = search_space.num_samples
    return search_space.__class__(**new_kwargs)


def derelativize_optimization_config_with_raw_status_quo(
    optimization_config: OptimizationConfig,
    modelbridge: "modelbridge_module.base.ModelBridge",
    observations: Optional[List[Observation]],
) -> OptimizationConfig:
    """Derelativize optimization_config using raw status-quo values"""
    tf = Derelativize(
        search_space=modelbridge.model_space.clone(),
        observations=observations,
        config={"use_raw_status_quo": True},
    )
    return tf.transform_optimization_config(
        optimization_config=optimization_config.clone(),
        modelbridge=modelbridge,
        fixed_features=ObservationFeatures(parameters={}),
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from logging import Logger
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
import torch
from ax.exceptions.core import SearchSpaceExhausted
from ax.models.base import Model
from ax.models.model_utils import (
    add_fixed_features,
    rejection_sample,
    tunable_feature_indices,
    validate_bounds,
)
from ax.models.types import TConfig
from ax.utils.common.docutils import copy_doc
from ax.utils.common.logger import get_logger
from ax.utils.common.typeutils import checked_cast, checked_cast_to_tuple
from botorch.utils.sampling import HitAndRunPolytopeSampler
from torch import Tensor


logger: Logger = get_logger(__name__)


class RandomModel(Model):
    """This class specifies the basic skeleton for a random model.

    As random generators do not make use of models, they do not implement
    the fit or predict methods.

    These models do not need data, or optimization configs.

    To satisfy search space parameter constraints, these models can use
    rejection sampling. To enable rejection sampling for a subclass, only
    only `_gen_samples` needs to be implemented, or alternatively,
    `_gen_unconstrained`/`gen` can be directly implemented.

    Attributes:
        deduplicate: If True (defaults to True), a single instantiation
            of the model will not return the same point twice. This flag
            is used in rejection sampling.
        seed: An optional seed value for scrambling.
        generated_points: A set of previously generated points to use
            for deduplication. These should be provided in the raw transformed
            space the model operates in.
        fallback_to_sample_polytope: If True, when rejection sampling fails,
            we fall back to the HitAndRunPolytopeSampler.
    """

    def __init__(
        self,
        deduplicate: bool = True,
        seed: Optional[int] = None,
        generated_points: Optional[np.ndarray] = None,
        fallback_to_sample_polytope: bool = False,
    ) -> None:
        super().__init__()
        self.deduplicate = deduplicate
        self.seed: int = (
            seed
            if seed is not None
            else checked_cast(int, torch.randint(high=100_000, size=(1,)).item())
        )
        # Used for deduplication.
        self.generated_points = generated_points
        self.fallback_to_sample_polytope = fallback_to_sample_polytope
        self.attempted_draws: int = 0

    def gen(
        self,
        n: int,
        bounds: List[Tuple[float, float]],
        linear_constraints: Optional[Tuple[np.ndarray, np.ndarray]] = None,
        fixed_features: Optional[Dict[int, float]] = None,
        model_gen_options: Optional[TConfig] = None,
        rounding_func: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Generate new candidates.

        Args:
            n: Number of candidates to generate.
            bounds: A list of (lower, upper) tuples for each column of X.
                Defined on [0, 1]^d.
            linear_constraints: A tuple of (A, b). For k linear constraints on
                d-dimensional x, A is (k x d) and b is (k x 1) such that
                A x <= b.
            fixed_features: A map {feature_index: value} for features that
                should be fixed to a particular value during generation.
            model_gen_options: A config dictionary that is passed along to the
                model.
            rounding_func: A function that rounds an optimization result
                appropriately (e.g., according to `round-trip` transformations).

        Returns:
            2-element tuple containing

            - (n x d) array of generated points.
            - Uniform weights, an n-array of ones for each point.

        """
        tf_indices = tunable_feature_indices(
            bounds=bounds, fixed_features=fixed_features
        )
        if fixed_features:
            fixed_feature_indices = np.array(list(fixed_features.keys()))
        else:
            fixed_feature_indices = np.array([])

        validate_bounds(bounds=bounds, fixed_feature_indices=fixed_feature_indices)
        attempted_draws = 0
        max_draws = None
        if model_gen_options:
            max_draws = model_gen_options.get("max_rs_draws")
            if max_draws is not None:
                max_draws = int(checked_cast_to_tuple((int, float), max_draws))
        try:
            # Always rejection sample, but this only rejects if there are
            # constraints or actual duplicates and deduplicate is specified.
            # If rejection sampling fails, fall back to polytope sampling
            points, attempted_draws = rejection_sample(
                gen_unconstrained=self._gen_unconstrained,
                n=n,
                d=len(bounds),
                tunable_feature_indices=tf_indices,
                linear_constraints=linear_constraints,
                deduplicate=self.deduplicate,
                max_draws=max_draws,
                fixed_features=fixed_features,
                rounding_func=rounding_func,
                existing_points=self.generated_points,
            )
        except SearchSpaceExhausted as e:
            if self.fallback_to_sample_polytope:
                logger.info(
                    "Rejection sampling exceeded specified maximum draws. "
                    "Falling back on HitAndRunPolytopeSampler instead of "
                    f"{self.__class__.__name__}."
                )
                # If rejection sampling fails, try polytope sampler.
                num_generated = (
                    len(self.generated_points)
                    if self.generated_points is not None
                    else 0
                )
                polytope_sampler = HitAndRunPolytopeSampler(
                    inequality_constraints=self._convert_inequality_constraints(
                        linear_constraints,
                    ),
                    equality_constraints=self._convert_equality_constraints(
                        d=len(bounds),
                        fixed_features=fixed_features,
                    ),
                    bounds=self._convert_bounds(bounds),
                    interior_point=self._get_last_point(),
                    n_burnin=100,
                    n_thinning=20,
                    seed=self.seed + num_generated,
                )
                points = polytope_sampler.draw(n=n).numpy()
                # TODO: Should this round & deduplicate?
            else:
                raise e

        self.attempted_draws = attempted_draws
        if self.deduplicate:
            if self.generated_points is None:
                self.generated_points = points
            else:
                self.generated_points = np.vstack([self.generated_points, points])
        return points, np.ones(len(points))

    @copy_doc(Model._get_state)
    def _get_state(self) -> Dict[str, Any]:
        state = super()._get_state()
        if not self.deduplicate:
            return state
        state.update({"generated_points": self.generated_points})
        return state

    def _gen_unconstrained(
        self,
        n: int,
        d: int,
        tunable_feature_indices: np.ndarray,
        fixed_features: Optional[Dict[int, float]] = None,
    ) -> np.ndarray:
        """Generate n points, from an unconstrained parameter space, using _gen_samples.

        Args:
            n: Number of points to generate.
            d: Dimension of parameter space.
            tunable_feature_indices: Parameter indices (in d) which are tunable.
            fixed_features: A map {feature_index: value} for features that
                should be fixed to a particular value during generation.

        Returns:
            An (n x d) array of generated points.

        """
        tunable_points = self._gen_samples(n=n, tunable_d=len(tunable_feature_indices))
        points = add_fixed_features(
            tunable_points=tunable_points,
            d=d,
            tunable_feature_indices=tunable_feature_indices,
            fixed_features=fixed_features,
        )
        return points

    def _gen_samples(self, n: int, tunable_d: int) -> np.ndarray:
        """Generate n samples on [0, 1]^d.

        Args:
            n: Number of points to generate.

        Returns:
            (n x d) array of generated points.

        """
        raise NotImplementedError("Base RandomModel can't generate samples.")

    def _convert_inequality_constraints(
        self, linear_constraints: Optional[Tuple[np.ndarray, np.ndarray]]
    ) -> Optional[Tuple[Tensor, Tensor]]:
        """Helper method to convert inequality constraints used by the rejection
        sampler to the format required for the polytope sampler.

            Args:
                linear_constraints: A tuple of (A, b). For k linear constraints on
                    d-dimensional x, A is (k x d) and b is (k x 1) such that
                    A x <= b.

            Returns:
                Optional 2-element tuple containing A and b as tensors
        """
        if linear_constraints is None:
            return None
        else:
            A = torch.as_tensor(linear_constraints[0], dtype=torch.double)
            b = torch.as_tensor(linear_constraints[1], dtype=torch.double)
            return A, b

    def _convert_equality_constraints(
        self, d: int, fixed_features: Optional[Dict[int, float]]
    ) -> Optional[Tuple[Tensor, Tensor]]:
        """Helper method to convert the fixed feature dictionary used by the rejection
        sampler to the corresponding matrix representation required for the polytope
        sampler.

            Args:
                d: dimension of samples
                fixed_features: A map {feature_index: value} for features that
                    should be fixed to a particular value during generation.

            Returns:
                Optional 2-element tuple containing C and c such that the equality
                constraints are defined by Cx = c
        """
        if fixed_features is None:
            return None

        n = len(fixed_features)
        fixed_indices = sorted(fixed_features.keys())
        fixed_vals = torch.tensor(
            [fixed_features[i] for i in fixed_indices], dtype=torch.double
        )
        constraint_matrix = torch.zeros((n, d), dtype=torch.double)
        for index in range(0, len(fixed_vals)):
            constraint_matrix[index, fixed_indices[index]] = 1.0
        return constraint_matrix, fixed_vals

    def _convert_bounds(self, bounds: List[Tuple[float, float]]) -> Optional[Tensor]:
        """Helper method to convert bounds list used by the rejectionsampler to the
        tensor format required for the polytope sampler.

            Args:
                bounds: A list of (lower, upper) tuples for each column of X.
                    Defined on [0, 1]^d.

            Returns:
                Optional 2 x d tensor representing the bounds
        """
        if bounds is None:
            return None
        else:
            return torch.tensor(bounds, dtype=torch.double).transpose(-1, -2)

    def _get_last_point(self) -> Optional[Tensor]:
        # Return the last sampled point when points have been sampled
        if self.generated_points is None:
            return None
        else:
            last_point = self.generated_points[-1, :].reshape((-1, 1))
            return torch.from_numpy(last_point).double()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from typing import Any, Dict, List, Optional, Set, Tuple

import numpy as np
from ax.core.observation import ObservationFeatures
from ax.modelbridge import ModelBridge


def predict_at_point(
    model: ModelBridge,
    obsf: ObservationFeatures,
    metric_names: Set[str],
    scalarized_metric_config: Optional[List[Dict[str, Any]]] = None,
) -> Tuple[Dict[str, float], Dict[str, float]]:
    """Make a prediction at a point.

    Returns mean and standard deviation in format expected by plotting.

    Args:
        model: ModelBridge
        obsf: ObservationFeatures for which to predict
        metric_names: Limit predictions to these metrics.
        scalarized_metric_config: An optional list of dicts specifying how to aggregate
            multiple metrics into a single scalarized metric. For each dict, the key is
            the name of the new scalarized metric, and the value is a dictionary mapping
            each metric to its weight. e.g.
            {"name": "metric1:agg", "weight": {"metric1_c1": 0.5, "metric1_c2": 0.5}}.

    Returns:
        A tuple containing

        - Map from metric name to prediction.
        - Map from metric name to standard error.
    """
    f_pred, cov_pred = model.predict([obsf])
    mean_pred_dict = {metric_name: pred[0] for metric_name, pred in f_pred.items()}
    cov_pred_dict = {}
    for metric_name1, pred_dict in cov_pred.items():
        cov_pred_dict[metric_name1] = {}
        for metric_name2, pred in pred_dict.items():
            cov_pred_dict[metric_name1][metric_name2] = pred[0]

    y_hat = {}
    se_hat = {}
    for metric_name in f_pred:
        if metric_name in metric_names:
            y_hat[metric_name] = mean_pred_dict[metric_name]
            se_hat[metric_name] = np.sqrt(cov_pred_dict[metric_name][metric_name])
    if scalarized_metric_config is not None:
        for agg_metric in scalarized_metric_config:
            agg_metric_name = agg_metric["name"]
            if agg_metric_name in metric_names:
                agg_metric_weight_dict = agg_metric["weight"]
                pred_mean, pred_var = _compute_scalarized_outcome(
                    mean_dict=mean_pred_dict,
                    cov_dict=cov_pred_dict,
                    agg_metric_weight_dict=agg_metric_weight_dict,
                )
                y_hat[agg_metric_name] = pred_mean
                se_hat[agg_metric_name] = np.sqrt(pred_var)
    return y_hat, se_hat


def predict_by_features(
    model: ModelBridge,
    label_to_feature_dict: Dict[int, ObservationFeatures],
    metric_names: Set[str],
) -> Dict[int, Dict[str, Tuple[float, float]]]:
    """Predict for given data points and model.

    Args:
        model: Model to be used for the prediction
        metric_names: Names of the metrics, for which to retrieve predictions.
        label_to_feature_dict: Mapping from an int label to
            a Parameterization. These data points are predicted.

    Returns:
        A mapping from an int label to a mapping of metric names to tuples
        of predicted metric mean and SEM, of form:
        { trial_index -> { metric_name: ( mean, SEM ) } }.
    """
    predictions_dict = {}  # Store predictions to return
    for label in label_to_feature_dict:
        try:
            y_hat, se_hat = predict_at_point(
                model=model,
                obsf=label_to_feature_dict[label],
                metric_names=metric_names,
            )
        except NotImplementedError:
            raise NotImplementedError(
                "The model associated with the current generation strategy "
                "step is not one that can be used for predicting values. "
                "For example, this may be the Sobol generator associated with the "
                "initialization step where quasi-random points are generated. "
                "Try again by calling the `AxClient.create_experiment()` "
                "method with the `choose_generation_strategy_kwargs="
                '{"num_initialization_trials": 0}` parameter if you are looking '
                "to use a generation strategy without an initialization step that "
                "proceeds straight to the Bayesian optimization step, but note "
                "that performance of Bayesian optimization can be suboptimal if "
                "search space is not sampled well in the initialization phase."
            )

        predictions_dict[label] = {
            metric: (
                y_hat[metric],
                se_hat[metric],
            )
            for metric in metric_names
        }

    return predictions_dict


def _compute_scalarized_outcome(
    mean_dict: Dict[str, float],
    cov_dict: Dict[str, Dict[str, float]],
    agg_metric_weight_dict: Dict[str, float],
) -> Tuple[float, float]:
    """Compute the mean and variance of a scalarized outcome.

    Args:
        mean_dict: Dictionary of means of individual metrics. e.g.
            {"metric1": 0.1, "metric2": 0.2}
        cov_dict: Dictionary of covariances of each metric pair. e.g.
            {"metric1": {"metric1": 0.25, "metric2": 0.0}, "metric2": {"metric1": 0.0,
            "metric2": 0.1}}
        agg_metric_weight_dict. Dictionary of scalarization weights of the scalarized
            outcome. e.g. {"name": "metric1:agg", "weight": {"metric1_c1": 0.5,
             "metric1_c2": 0.5}}
    """
    pred_mean = 0
    pred_var = 0
    component_metrics = list(agg_metric_weight_dict.keys())
    for i, metric_name in enumerate(component_metrics):
        weight = agg_metric_weight_dict[metric_name]
        pred_mean += weight * mean_dict[metric_name]
        pred_var += (weight**2) * cov_dict[metric_name][metric_name]
        # include cross-metric covariance
        for metric_name2 in component_metrics[(i + 1) :]:
            weight2 = agg_metric_weight_dict[metric_name2]
            pred_var += 2 * weight * weight2 * cov_dict[metric_name][metric_name2]
    return pred_mean, pred_var
